# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import some_ip_sd
class SomeIp(KaitaiStruct):
    """SOME/IP (Scalable service-Oriented MiddlewarE over IP) is an automotive/embedded
    communication protocol which supports remoteprocedure calls, event notifications
    and the underlying serialization/wire format.
    
    .. seealso::
       Source - https://www.autosar.org/fileadmin/user_upload/standards/foundation/19-11/AUTOSAR_PRS_SOMEIPProtocol.pdf
    """
    SEQ_FIELDS = ["header", "payload"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = SomeIp.Header(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['payload']['start'] = self._io.pos()
        _on = self.header.message_id.value
        if _on == 4294934784:
            self._raw_payload = self._io.read_bytes((self.header.length - 8))
            _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
            self.payload = some_ip_sd.SomeIpSd(_io__raw_payload)
            self.payload._read()
        else:
            self.payload = self._io.read_bytes((self.header.length - 8))
        self._debug['payload']['end'] = self._io.pos()

    class Header(KaitaiStruct):

        class MessageTypeEnum(Enum):
            request = 0
            request_no_return = 1
            notification = 2
            request_ack = 64
            request_no_return_ack = 65
            notification_ack = 66
            response = 128
            error = 129
            response_ack = 192
            error_ack = 193

        class ReturnCodeEnum(Enum):
            ok = 0
            not_ok = 1
            unknown_service = 2
            unknown_method = 3
            not_ready = 4
            not_reachable = 5
            time_out = 6
            wrong_protocol_version = 7
            wrong_interface_version = 8
            malformed_message = 9
            wrong_message_type = 10
        SEQ_FIELDS = ["message_id", "length", "request_id", "protocol_version", "interface_version", "message_type", "return_code"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['message_id']['start'] = self._io.pos()
            self._raw_message_id = self._io.read_bytes(4)
            _io__raw_message_id = KaitaiStream(BytesIO(self._raw_message_id))
            self.message_id = SomeIp.Header.MessageId(_io__raw_message_id, self, self._root)
            self.message_id._read()
            self._debug['message_id']['end'] = self._io.pos()
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u4be()
            self._debug['length']['end'] = self._io.pos()
            self._debug['request_id']['start'] = self._io.pos()
            self._raw_request_id = self._io.read_bytes(4)
            _io__raw_request_id = KaitaiStream(BytesIO(self._raw_request_id))
            self.request_id = SomeIp.Header.RequestId(_io__raw_request_id, self, self._root)
            self.request_id._read()
            self._debug['request_id']['end'] = self._io.pos()
            self._debug['protocol_version']['start'] = self._io.pos()
            self.protocol_version = self._io.read_u1()
            self._debug['protocol_version']['end'] = self._io.pos()
            self._debug['interface_version']['start'] = self._io.pos()
            self.interface_version = self._io.read_u1()
            self._debug['interface_version']['end'] = self._io.pos()
            self._debug['message_type']['start'] = self._io.pos()
            self.message_type = KaitaiStream.resolve_enum(SomeIp.Header.MessageTypeEnum, self._io.read_u1())
            self._debug['message_type']['end'] = self._io.pos()
            self._debug['return_code']['start'] = self._io.pos()
            self.return_code = KaitaiStream.resolve_enum(SomeIp.Header.ReturnCodeEnum, self._io.read_u1())
            self._debug['return_code']['end'] = self._io.pos()

        class MessageId(KaitaiStruct):
            """[PRS_SOMEIP_00035] The assignment of the Message ID shall be up to
            the user. However, the Message ID shall be unique for the whole
            system (i.e. the vehicle). TheMessage ID is similar to a CAN ID and
            should be handled via a comparable process.
            [PRS_SOMEIP_00038] Message IDs of method calls shall be structured in
            the ID with 2^16 services with 2^15 methods.
            
            .. seealso::
               AUTOSAR_PRS_SOMEIPProtocol.pdf 4.1.1.1  Message ID
            """
            SEQ_FIELDS = ["service_id", "sub_id", "method_id", "event_id"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['service_id']['start'] = self._io.pos()
                self.service_id = self._io.read_u2be()
                self._debug['service_id']['end'] = self._io.pos()
                self._debug['sub_id']['start'] = self._io.pos()
                self.sub_id = self._io.read_bits_int_be(1) != 0
                self._debug['sub_id']['end'] = self._io.pos()
                if self.sub_id == False:
                    self._debug['method_id']['start'] = self._io.pos()
                    self.method_id = self._io.read_bits_int_be(15)
                    self._debug['method_id']['end'] = self._io.pos()

                if self.sub_id == True:
                    self._debug['event_id']['start'] = self._io.pos()
                    self.event_id = self._io.read_bits_int_be(15)
                    self._debug['event_id']['end'] = self._io.pos()


            @property
            def value(self):
                """The value provides the undissected Message ID."""
                if hasattr(self, '_m_value'):
                    return self._m_value if hasattr(self, '_m_value') else None

                _pos = self._io.pos()
                self._io.seek(0)
                self._debug['_m_value']['start'] = self._io.pos()
                self._m_value = self._io.read_u4be()
                self._debug['_m_value']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_value if hasattr(self, '_m_value') else None


        class RequestId(KaitaiStruct):
            """The Request ID allows a provider and subscriber to differentiate
            multiple parallel usesof the same method, event, getter or setter.
            
            .. seealso::
               AUTOSAR_PRS_SOMEIPProtocol.pdf - section 4.1.1.3  Request ID
            """
            SEQ_FIELDS = ["client_id", "session_id"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['client_id']['start'] = self._io.pos()
                self.client_id = self._io.read_u2be()
                self._debug['client_id']['end'] = self._io.pos()
                self._debug['session_id']['start'] = self._io.pos()
                self.session_id = self._io.read_u2be()
                self._debug['session_id']['end'] = self._io.pos()

            @property
            def value(self):
                """The value provides the undissected Request ID."""
                if hasattr(self, '_m_value'):
                    return self._m_value if hasattr(self, '_m_value') else None

                _pos = self._io.pos()
                self._io.seek(0)
                self._debug['_m_value']['start'] = self._io.pos()
                self._m_value = self._io.read_u4be()
                self._debug['_m_value']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_value if hasattr(self, '_m_value') else None


        @property
        def is_valid_service_discovery(self):
            """auxillary value.
            
            .. seealso::
               AUTOSAR_PRS_SOMEIPServiceDiscoveryProtocol.pdf - section 4.1.2.1 General Requirements
            """
            if hasattr(self, '_m_is_valid_service_discovery'):
                return self._m_is_valid_service_discovery if hasattr(self, '_m_is_valid_service_discovery') else None

            self._m_is_valid_service_discovery =  ((self.message_id.value == 4294934784) and (self.protocol_version == 1) and (self.interface_version == 1) and (self.message_type == SomeIp.Header.MessageTypeEnum.notification) and (self.return_code == SomeIp.Header.ReturnCodeEnum.ok)) 
            return self._m_is_valid_service_discovery if hasattr(self, '_m_is_valid_service_discovery') else None






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class GptPartitionTable(KaitaiStruct):
    """
    .. seealso::
       Source - https://en.wikipedia.org/wiki/GUID_Partition_Table
    """
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class PartitionEntry(KaitaiStruct):
        SEQ_FIELDS = ["type_guid", "guid", "first_lba", "last_lba", "attributes", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type_guid']['start'] = self._io.pos()
            self.type_guid = self._io.read_bytes(16)
            self._debug['type_guid']['end'] = self._io.pos()
            self._debug['guid']['start'] = self._io.pos()
            self.guid = self._io.read_bytes(16)
            self._debug['guid']['end'] = self._io.pos()
            self._debug['first_lba']['start'] = self._io.pos()
            self.first_lba = self._io.read_u8le()
            self._debug['first_lba']['end'] = self._io.pos()
            self._debug['last_lba']['start'] = self._io.pos()
            self.last_lba = self._io.read_u8le()
            self._debug['last_lba']['end'] = self._io.pos()
            self._debug['attributes']['start'] = self._io.pos()
            self.attributes = self._io.read_u8le()
            self._debug['attributes']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (self._io.read_bytes(72)).decode(u"UTF-16LE")
            self._debug['name']['end'] = self._io.pos()


    class PartitionHeader(KaitaiStruct):
        SEQ_FIELDS = ["signature", "revision", "header_size", "crc32_header", "reserved", "current_lba", "backup_lba", "first_usable_lba", "last_usable_lba", "disk_guid", "entries_start", "entries_count", "entries_size", "crc32_array"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['signature']['start'] = self._io.pos()
            self.signature = self._io.read_bytes(8)
            self._debug['signature']['end'] = self._io.pos()
            if not self.signature == b"\x45\x46\x49\x20\x50\x41\x52\x54":
                raise kaitaistruct.ValidationNotEqualError(b"\x45\x46\x49\x20\x50\x41\x52\x54", self.signature, self._io, u"/types/partition_header/seq/0")
            self._debug['revision']['start'] = self._io.pos()
            self.revision = self._io.read_u4le()
            self._debug['revision']['end'] = self._io.pos()
            self._debug['header_size']['start'] = self._io.pos()
            self.header_size = self._io.read_u4le()
            self._debug['header_size']['end'] = self._io.pos()
            self._debug['crc32_header']['start'] = self._io.pos()
            self.crc32_header = self._io.read_u4le()
            self._debug['crc32_header']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_u4le()
            self._debug['reserved']['end'] = self._io.pos()
            self._debug['current_lba']['start'] = self._io.pos()
            self.current_lba = self._io.read_u8le()
            self._debug['current_lba']['end'] = self._io.pos()
            self._debug['backup_lba']['start'] = self._io.pos()
            self.backup_lba = self._io.read_u8le()
            self._debug['backup_lba']['end'] = self._io.pos()
            self._debug['first_usable_lba']['start'] = self._io.pos()
            self.first_usable_lba = self._io.read_u8le()
            self._debug['first_usable_lba']['end'] = self._io.pos()
            self._debug['last_usable_lba']['start'] = self._io.pos()
            self.last_usable_lba = self._io.read_u8le()
            self._debug['last_usable_lba']['end'] = self._io.pos()
            self._debug['disk_guid']['start'] = self._io.pos()
            self.disk_guid = self._io.read_bytes(16)
            self._debug['disk_guid']['end'] = self._io.pos()
            self._debug['entries_start']['start'] = self._io.pos()
            self.entries_start = self._io.read_u8le()
            self._debug['entries_start']['end'] = self._io.pos()
            self._debug['entries_count']['start'] = self._io.pos()
            self.entries_count = self._io.read_u4le()
            self._debug['entries_count']['end'] = self._io.pos()
            self._debug['entries_size']['start'] = self._io.pos()
            self.entries_size = self._io.read_u4le()
            self._debug['entries_size']['end'] = self._io.pos()
            self._debug['crc32_array']['start'] = self._io.pos()
            self.crc32_array = self._io.read_u4le()
            self._debug['crc32_array']['end'] = self._io.pos()

        @property
        def entries(self):
            if hasattr(self, '_m_entries'):
                return self._m_entries if hasattr(self, '_m_entries') else None

            io = self._root._io
            _pos = io.pos()
            io.seek((self.entries_start * self._root.sector_size))
            self._debug['_m_entries']['start'] = io.pos()
            self._raw__m_entries = [None] * (self.entries_count)
            self._m_entries = [None] * (self.entries_count)
            for i in range(self.entries_count):
                if not 'arr' in self._debug['_m_entries']:
                    self._debug['_m_entries']['arr'] = []
                self._debug['_m_entries']['arr'].append({'start': io.pos()})
                self._raw__m_entries[i] = io.read_bytes(self.entries_size)
                _io__raw__m_entries = KaitaiStream(BytesIO(self._raw__m_entries[i]))
                _t__m_entries = GptPartitionTable.PartitionEntry(_io__raw__m_entries, self, self._root)
                _t__m_entries._read()
                self._m_entries[i] = _t__m_entries
                self._debug['_m_entries']['arr'][i]['end'] = io.pos()

            self._debug['_m_entries']['end'] = io.pos()
            io.seek(_pos)
            return self._m_entries if hasattr(self, '_m_entries') else None


    @property
    def sector_size(self):
        if hasattr(self, '_m_sector_size'):
            return self._m_sector_size if hasattr(self, '_m_sector_size') else None

        self._m_sector_size = 512
        return self._m_sector_size if hasattr(self, '_m_sector_size') else None

    @property
    def primary(self):
        if hasattr(self, '_m_primary'):
            return self._m_primary if hasattr(self, '_m_primary') else None

        io = self._root._io
        _pos = io.pos()
        io.seek(self._root.sector_size)
        self._debug['_m_primary']['start'] = io.pos()
        self._m_primary = GptPartitionTable.PartitionHeader(io, self, self._root)
        self._m_primary._read()
        self._debug['_m_primary']['end'] = io.pos()
        io.seek(_pos)
        return self._m_primary if hasattr(self, '_m_primary') else None

    @property
    def backup(self):
        if hasattr(self, '_m_backup'):
            return self._m_backup if hasattr(self, '_m_backup') else None

        io = self._root._io
        _pos = io.pos()
        io.seek((self._io.size() - self._root.sector_size))
        self._debug['_m_backup']['start'] = io.pos()
        self._m_backup = GptPartitionTable.PartitionHeader(io, self, self._root)
        self._m_backup._read()
        self._debug['_m_backup']['end'] = io.pos()
        io.seek(_pos)
        return self._m_backup if hasattr(self, '_m_backup') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import vlq_base128_be
class Sqlite3(KaitaiStruct):
    """SQLite3 is a popular serverless SQL engine, implemented as a library
    to be used within other applications. It keeps its databases as
    regular disk files.
    
    Every database file is segmented into pages. First page (starting at
    the very beginning) is special: it contains a file-global header
    which specifies some data relevant to proper parsing (i.e. format
    versions, size of page, etc). After the header, normal contents of
    the first page follow.
    
    Each page would be of some type, and generally, they would be
    reached via the links starting from the first page. First page type
    (`root_page`) is always "btree_page".
    
    .. seealso::
       Source - https://www.sqlite.org/fileformat.html
    """

    class Versions(Enum):
        legacy = 1
        wal = 2

    class Encodings(Enum):
        utf_8 = 1
        utf_16le = 2
        utf_16be = 3
    SEQ_FIELDS = ["magic", "len_page_mod", "write_version", "read_version", "reserved_space", "max_payload_frac", "min_payload_frac", "leaf_payload_frac", "file_change_counter", "num_pages", "first_freelist_trunk_page", "num_freelist_pages", "schema_cookie", "schema_format", "def_page_cache_size", "largest_root_page", "text_encoding", "user_version", "is_incremental_vacuum", "application_id", "reserved", "version_valid_for", "sqlite_version_number", "root_page"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(16)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x53\x51\x4C\x69\x74\x65\x20\x66\x6F\x72\x6D\x61\x74\x20\x33\x00":
            raise kaitaistruct.ValidationNotEqualError(b"\x53\x51\x4C\x69\x74\x65\x20\x66\x6F\x72\x6D\x61\x74\x20\x33\x00", self.magic, self._io, u"/seq/0")
        self._debug['len_page_mod']['start'] = self._io.pos()
        self.len_page_mod = self._io.read_u2be()
        self._debug['len_page_mod']['end'] = self._io.pos()
        self._debug['write_version']['start'] = self._io.pos()
        self.write_version = KaitaiStream.resolve_enum(Sqlite3.Versions, self._io.read_u1())
        self._debug['write_version']['end'] = self._io.pos()
        self._debug['read_version']['start'] = self._io.pos()
        self.read_version = KaitaiStream.resolve_enum(Sqlite3.Versions, self._io.read_u1())
        self._debug['read_version']['end'] = self._io.pos()
        self._debug['reserved_space']['start'] = self._io.pos()
        self.reserved_space = self._io.read_u1()
        self._debug['reserved_space']['end'] = self._io.pos()
        self._debug['max_payload_frac']['start'] = self._io.pos()
        self.max_payload_frac = self._io.read_u1()
        self._debug['max_payload_frac']['end'] = self._io.pos()
        self._debug['min_payload_frac']['start'] = self._io.pos()
        self.min_payload_frac = self._io.read_u1()
        self._debug['min_payload_frac']['end'] = self._io.pos()
        self._debug['leaf_payload_frac']['start'] = self._io.pos()
        self.leaf_payload_frac = self._io.read_u1()
        self._debug['leaf_payload_frac']['end'] = self._io.pos()
        self._debug['file_change_counter']['start'] = self._io.pos()
        self.file_change_counter = self._io.read_u4be()
        self._debug['file_change_counter']['end'] = self._io.pos()
        self._debug['num_pages']['start'] = self._io.pos()
        self.num_pages = self._io.read_u4be()
        self._debug['num_pages']['end'] = self._io.pos()
        self._debug['first_freelist_trunk_page']['start'] = self._io.pos()
        self.first_freelist_trunk_page = self._io.read_u4be()
        self._debug['first_freelist_trunk_page']['end'] = self._io.pos()
        self._debug['num_freelist_pages']['start'] = self._io.pos()
        self.num_freelist_pages = self._io.read_u4be()
        self._debug['num_freelist_pages']['end'] = self._io.pos()
        self._debug['schema_cookie']['start'] = self._io.pos()
        self.schema_cookie = self._io.read_u4be()
        self._debug['schema_cookie']['end'] = self._io.pos()
        self._debug['schema_format']['start'] = self._io.pos()
        self.schema_format = self._io.read_u4be()
        self._debug['schema_format']['end'] = self._io.pos()
        self._debug['def_page_cache_size']['start'] = self._io.pos()
        self.def_page_cache_size = self._io.read_u4be()
        self._debug['def_page_cache_size']['end'] = self._io.pos()
        self._debug['largest_root_page']['start'] = self._io.pos()
        self.largest_root_page = self._io.read_u4be()
        self._debug['largest_root_page']['end'] = self._io.pos()
        self._debug['text_encoding']['start'] = self._io.pos()
        self.text_encoding = KaitaiStream.resolve_enum(Sqlite3.Encodings, self._io.read_u4be())
        self._debug['text_encoding']['end'] = self._io.pos()
        self._debug['user_version']['start'] = self._io.pos()
        self.user_version = self._io.read_u4be()
        self._debug['user_version']['end'] = self._io.pos()
        self._debug['is_incremental_vacuum']['start'] = self._io.pos()
        self.is_incremental_vacuum = self._io.read_u4be()
        self._debug['is_incremental_vacuum']['end'] = self._io.pos()
        self._debug['application_id']['start'] = self._io.pos()
        self.application_id = self._io.read_u4be()
        self._debug['application_id']['end'] = self._io.pos()
        self._debug['reserved']['start'] = self._io.pos()
        self.reserved = self._io.read_bytes(20)
        self._debug['reserved']['end'] = self._io.pos()
        self._debug['version_valid_for']['start'] = self._io.pos()
        self.version_valid_for = self._io.read_u4be()
        self._debug['version_valid_for']['end'] = self._io.pos()
        self._debug['sqlite_version_number']['start'] = self._io.pos()
        self.sqlite_version_number = self._io.read_u4be()
        self._debug['sqlite_version_number']['end'] = self._io.pos()
        self._debug['root_page']['start'] = self._io.pos()
        self.root_page = Sqlite3.BtreePage(self._io, self, self._root)
        self.root_page._read()
        self._debug['root_page']['end'] = self._io.pos()

    class Serial(KaitaiStruct):
        SEQ_FIELDS = ["code"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = vlq_base128_be.VlqBase128Be(self._io)
            self.code._read()
            self._debug['code']['end'] = self._io.pos()

        @property
        def is_blob(self):
            if hasattr(self, '_m_is_blob'):
                return self._m_is_blob if hasattr(self, '_m_is_blob') else None

            self._m_is_blob =  ((self.code.value >= 12) and ((self.code.value % 2) == 0)) 
            return self._m_is_blob if hasattr(self, '_m_is_blob') else None

        @property
        def is_string(self):
            if hasattr(self, '_m_is_string'):
                return self._m_is_string if hasattr(self, '_m_is_string') else None

            self._m_is_string =  ((self.code.value >= 13) and ((self.code.value % 2) == 1)) 
            return self._m_is_string if hasattr(self, '_m_is_string') else None

        @property
        def len_content(self):
            if hasattr(self, '_m_len_content'):
                return self._m_len_content if hasattr(self, '_m_len_content') else None

            if self.code.value >= 12:
                self._m_len_content = (self.code.value - 12) // 2

            return self._m_len_content if hasattr(self, '_m_len_content') else None


    class BtreePage(KaitaiStruct):
        SEQ_FIELDS = ["page_type", "first_freeblock", "num_cells", "ofs_cells", "num_frag_free_bytes", "right_ptr", "cells"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['page_type']['start'] = self._io.pos()
            self.page_type = self._io.read_u1()
            self._debug['page_type']['end'] = self._io.pos()
            self._debug['first_freeblock']['start'] = self._io.pos()
            self.first_freeblock = self._io.read_u2be()
            self._debug['first_freeblock']['end'] = self._io.pos()
            self._debug['num_cells']['start'] = self._io.pos()
            self.num_cells = self._io.read_u2be()
            self._debug['num_cells']['end'] = self._io.pos()
            self._debug['ofs_cells']['start'] = self._io.pos()
            self.ofs_cells = self._io.read_u2be()
            self._debug['ofs_cells']['end'] = self._io.pos()
            self._debug['num_frag_free_bytes']['start'] = self._io.pos()
            self.num_frag_free_bytes = self._io.read_u1()
            self._debug['num_frag_free_bytes']['end'] = self._io.pos()
            if  ((self.page_type == 2) or (self.page_type == 5)) :
                self._debug['right_ptr']['start'] = self._io.pos()
                self.right_ptr = self._io.read_u4be()
                self._debug['right_ptr']['end'] = self._io.pos()

            self._debug['cells']['start'] = self._io.pos()
            self.cells = [None] * (self.num_cells)
            for i in range(self.num_cells):
                if not 'arr' in self._debug['cells']:
                    self._debug['cells']['arr'] = []
                self._debug['cells']['arr'].append({'start': self._io.pos()})
                _t_cells = Sqlite3.RefCell(self._io, self, self._root)
                _t_cells._read()
                self.cells[i] = _t_cells
                self._debug['cells']['arr'][i]['end'] = self._io.pos()

            self._debug['cells']['end'] = self._io.pos()


    class CellIndexLeaf(KaitaiStruct):
        """
        .. seealso::
           Source - https://www.sqlite.org/fileformat.html#b_tree_pages
        """
        SEQ_FIELDS = ["len_payload", "payload"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_payload']['start'] = self._io.pos()
            self.len_payload = vlq_base128_be.VlqBase128Be(self._io)
            self.len_payload._read()
            self._debug['len_payload']['end'] = self._io.pos()
            self._debug['payload']['start'] = self._io.pos()
            self._raw_payload = self._io.read_bytes(self.len_payload.value)
            _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
            self.payload = Sqlite3.CellPayload(_io__raw_payload, self, self._root)
            self.payload._read()
            self._debug['payload']['end'] = self._io.pos()


    class Serials(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = vlq_base128_be.VlqBase128Be(self._io)
                _t_entries._read()
                self.entries.append(_t_entries)
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()


    class CellTableLeaf(KaitaiStruct):
        """
        .. seealso::
           Source - https://www.sqlite.org/fileformat.html#b_tree_pages
        """
        SEQ_FIELDS = ["len_payload", "row_id", "payload"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_payload']['start'] = self._io.pos()
            self.len_payload = vlq_base128_be.VlqBase128Be(self._io)
            self.len_payload._read()
            self._debug['len_payload']['end'] = self._io.pos()
            self._debug['row_id']['start'] = self._io.pos()
            self.row_id = vlq_base128_be.VlqBase128Be(self._io)
            self.row_id._read()
            self._debug['row_id']['end'] = self._io.pos()
            self._debug['payload']['start'] = self._io.pos()
            self._raw_payload = self._io.read_bytes(self.len_payload.value)
            _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
            self.payload = Sqlite3.CellPayload(_io__raw_payload, self, self._root)
            self.payload._read()
            self._debug['payload']['end'] = self._io.pos()


    class CellPayload(KaitaiStruct):
        """
        .. seealso::
           Source - https://sqlite.org/fileformat2.html#record_format
        """
        SEQ_FIELDS = ["len_header_and_len", "column_serials", "column_contents"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_header_and_len']['start'] = self._io.pos()
            self.len_header_and_len = vlq_base128_be.VlqBase128Be(self._io)
            self.len_header_and_len._read()
            self._debug['len_header_and_len']['end'] = self._io.pos()
            self._debug['column_serials']['start'] = self._io.pos()
            self._raw_column_serials = self._io.read_bytes((self.len_header_and_len.value - 1))
            _io__raw_column_serials = KaitaiStream(BytesIO(self._raw_column_serials))
            self.column_serials = Sqlite3.Serials(_io__raw_column_serials, self, self._root)
            self.column_serials._read()
            self._debug['column_serials']['end'] = self._io.pos()
            self._debug['column_contents']['start'] = self._io.pos()
            self.column_contents = [None] * (len(self.column_serials.entries))
            for i in range(len(self.column_serials.entries)):
                if not 'arr' in self._debug['column_contents']:
                    self._debug['column_contents']['arr'] = []
                self._debug['column_contents']['arr'].append({'start': self._io.pos()})
                _t_column_contents = Sqlite3.ColumnContent(self.column_serials.entries[i], self._io, self, self._root)
                _t_column_contents._read()
                self.column_contents[i] = _t_column_contents
                self._debug['column_contents']['arr'][i]['end'] = self._io.pos()

            self._debug['column_contents']['end'] = self._io.pos()


    class CellTableInterior(KaitaiStruct):
        """
        .. seealso::
           Source - https://www.sqlite.org/fileformat.html#b_tree_pages
        """
        SEQ_FIELDS = ["left_child_page", "row_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['left_child_page']['start'] = self._io.pos()
            self.left_child_page = self._io.read_u4be()
            self._debug['left_child_page']['end'] = self._io.pos()
            self._debug['row_id']['start'] = self._io.pos()
            self.row_id = vlq_base128_be.VlqBase128Be(self._io)
            self.row_id._read()
            self._debug['row_id']['end'] = self._io.pos()


    class CellIndexInterior(KaitaiStruct):
        """
        .. seealso::
           Source - https://www.sqlite.org/fileformat.html#b_tree_pages
        """
        SEQ_FIELDS = ["left_child_page", "len_payload", "payload"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['left_child_page']['start'] = self._io.pos()
            self.left_child_page = self._io.read_u4be()
            self._debug['left_child_page']['end'] = self._io.pos()
            self._debug['len_payload']['start'] = self._io.pos()
            self.len_payload = vlq_base128_be.VlqBase128Be(self._io)
            self.len_payload._read()
            self._debug['len_payload']['end'] = self._io.pos()
            self._debug['payload']['start'] = self._io.pos()
            self._raw_payload = self._io.read_bytes(self.len_payload.value)
            _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
            self.payload = Sqlite3.CellPayload(_io__raw_payload, self, self._root)
            self.payload._read()
            self._debug['payload']['end'] = self._io.pos()


    class ColumnContent(KaitaiStruct):
        SEQ_FIELDS = ["as_int", "as_float", "as_blob", "as_str"]
        def __init__(self, ser, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.ser = ser
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if  ((self.serial_type.code.value >= 1) and (self.serial_type.code.value <= 6)) :
                self._debug['as_int']['start'] = self._io.pos()
                _on = self.serial_type.code.value
                if _on == 4:
                    self.as_int = self._io.read_u4be()
                elif _on == 6:
                    self.as_int = self._io.read_u8be()
                elif _on == 1:
                    self.as_int = self._io.read_u1()
                elif _on == 3:
                    self.as_int = self._io.read_bits_int_be(24)
                elif _on == 5:
                    self.as_int = self._io.read_bits_int_be(48)
                elif _on == 2:
                    self.as_int = self._io.read_u2be()
                self._debug['as_int']['end'] = self._io.pos()

            if self.serial_type.code.value == 7:
                self._debug['as_float']['start'] = self._io.pos()
                self.as_float = self._io.read_f8be()
                self._debug['as_float']['end'] = self._io.pos()

            if self.serial_type.is_blob:
                self._debug['as_blob']['start'] = self._io.pos()
                self.as_blob = self._io.read_bytes(self.serial_type.len_content)
                self._debug['as_blob']['end'] = self._io.pos()

            self._debug['as_str']['start'] = self._io.pos()
            self.as_str = (self._io.read_bytes(self.serial_type.len_content)).decode(u"UTF-8")
            self._debug['as_str']['end'] = self._io.pos()

        @property
        def serial_type(self):
            if hasattr(self, '_m_serial_type'):
                return self._m_serial_type if hasattr(self, '_m_serial_type') else None

            self._m_serial_type = self.ser
            return self._m_serial_type if hasattr(self, '_m_serial_type') else None


    class RefCell(KaitaiStruct):
        SEQ_FIELDS = ["ofs_body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs_body']['start'] = self._io.pos()
            self.ofs_body = self._io.read_u2be()
            self._debug['ofs_body']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek(self.ofs_body)
            self._debug['_m_body']['start'] = self._io.pos()
            _on = self._parent.page_type
            if _on == 13:
                self._m_body = Sqlite3.CellTableLeaf(self._io, self, self._root)
                self._m_body._read()
            elif _on == 5:
                self._m_body = Sqlite3.CellTableInterior(self._io, self, self._root)
                self._m_body._read()
            elif _on == 10:
                self._m_body = Sqlite3.CellIndexLeaf(self._io, self, self._root)
                self._m_body._read()
            elif _on == 2:
                self._m_body = Sqlite3.CellIndexInterior(self._io, self, self._root)
                self._m_body._read()
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    @property
    def len_page(self):
        if hasattr(self, '_m_len_page'):
            return self._m_len_page if hasattr(self, '_m_len_page') else None

        self._m_len_page = (65536 if self.len_page_mod == 1 else self.len_page_mod)
        return self._m_len_page if hasattr(self, '_m_len_page') else None





import base64
import math
import mimetypes
import os
import unicodedata

jinja2 = None


TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
TEMPLATE = None


def assign_ids(sbud):
    i = 0
    matches = list(sbud['struc'])
    stack = list(matches)
    while stack:
        m = stack.pop()
        assert 'uid' not in m
        m['uid'] = i
        i += 1
        stack += list(m['subEls'])
    return matches


def generate(file_path, sbud):
    global TEMPLATE, jinja2
    if jinja2 is None:
        # Dynamically load jinja2 at runtime so it is not an installation dependency for setup.py
        import jinja2 as j2
        jinja2 = j2

    if TEMPLATE is None:
        TEMPLATE = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATE_DIR)).get_template('template.html')

    matches = assign_ids(sbud)

    input_bytes = sbud['length']
    with open(file_path, 'rb') as input_file:
        class ReadUnicode():
            def __init__(self):
                self.reset = False

            def tell(self):
                if not self.reset:
                    return 0
                else:
                    return input_file.tell()

            def translate(self, b, monospace=True):
                if b is None or len(b) == 0 or b == b' ':
                    return '&nbsp;'
                elif b == b'\n':
                    if monospace:
                        return '\u2424'
                    else:
                        return '\u2424</span><br /><span>'
                elif b == b'\t':
                    if monospace:
                        return b'\xE2\xAD\xBE'
                    else:
                        return '\t'
                elif b == b'\r':
                    return '\u240d'
                try:
                    u = b.decode('utf-8')
                    if unicodedata.category(u) == 'Cc':
                        # This is a control character
                        return '\ufffd'
                    else:
                        return u
                except UnicodeDecodeError:
                    return '\ufffd'

            def __iter__(self):
                input_file.seek(0)
                i = 0
                while True:
                    b = input_file.read(1)
                    if b is None or len(b) < 1:
                        break
                    yield i, self.translate(b, monospace=False)
                    i += 1

            def __call__(self):
                if not self.reset:
                    input_file.seek(0)
                    self.reset = True
                b = input_file.read(1)
                return self.translate(b)

        mime_type = mimetypes.guess_type(file_path)[0]
        if mime_type is None:
            mime_type = 'application/octet-stream'

        def _decoded_matches(m=None):
            if m is None:
                return
            if 'decoded' in m:
                decoded = ''
                for b in base64.b64decode(m['decoded']).replace(b'\r\n', b'<br />') \
                        .replace(b'\n', b'<br />').replace(b'\r', b'\xE2\x8F\x8E'):
                    try:
                        decoded += bytes([b]).decode('utf-8')
                    except UnicodeDecodeError:
                        decoded += f"\\x{int(b)}"
                yield m, decoded
            if 'subEls' in m:
                for a in m['subEls']:
                    yield from _decoded_matches(a)

        def decoded_matches():
            for m in matches:
                yield from _decoded_matches(m)

        return TEMPLATE.render(
            filename=os.path.split(file_path)[-1],
            encoded=sbud['b64contents'],
            matches=matches,
            input_file=input_file,
            input_bytes=input_bytes,
            math=math,
            read_unicode=ReadUnicode(),
            mime_type=mime_type,
            decoded_matches=decoded_matches
        )


if __name__ == '__main__':
    import json
    import sys

    with open(sys.argv[2], 'r') as f:
        print(generate(sys.argv[1], json.load(f)))



#!/usr/bin/env python3

"""
This module automates compilation of Kaitai Struct definitions into Python code.

This script is called from PolyFile's setup.py to compile the entire Kaitai Struct format library at build time.
Therefore, this script should always be self-contained and not require any dependencies outside of the Python standard
library.

"""
import json
from pathlib import Path
import shutil
import subprocess
from typing import Iterable, List, Union


KAITAI_COMPILER_NAME: str = "kaitai-struct-compiler"


class KaitaiError(RuntimeError):
    pass


def has_kaitai_compiler() -> bool:
    return shutil.which(KAITAI_COMPILER_NAME) is not None


class CompiledKSY:
    def __init__(self, class_name: str, python_path: Union[str, Path], dependencies: Iterable["CompiledKSY"] = ()):
        self.class_name: str = class_name
        if not isinstance(python_path, Path):
            python_path = Path(python_path)
        self.python_path: Path = python_path
        self.dependencies: List[CompiledKSY] = list(dependencies)

    def __repr__(self):
        return f"{self.__class__.__name__}(class_name={self.class_name!r}, python_path={self.python_path!r}, "\
               f"dependencies={self.dependencies!r})"


def compile(ksy_path: Union[str, Path], output_directory: Union[str, Path]) -> CompiledKSY:
    """Returns the list of compiled KSYs; the original spec being first, followed by its dependencies"""
    if not has_kaitai_compiler():
        raise KaitaiError(f"{KAITAI_COMPILER_NAME} not found! Please make sure it is in your PATH")

    if not isinstance(output_directory, Path):
        output_directory = Path(output_directory)

    output_directory.mkdir(parents=True, exist_ok=True)

    cmd = [
        KAITAI_COMPILER_NAME, "--target", "python", "--outdir", str(output_directory), str(ksy_path),
        "--debug", "--ksc-json-output", "-I", Path.cwd(), "--python-package", "polyfile.kaitai.parsers"
    ]

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if err:
        raise KaitaiError(err.decode("utf-8"))
    elif proc.returncode != 0:
        raise KaitaiError(f"`{' '.join(cmd)}` returned with non-zero exit code {proc.returncode}")

    result = json.loads(out)

    if "errors" in result[ksy_path] and result[ksy_path]["errors"]:
        raise KaitaiError(f"Error compiling {ksy_path}: {result[ksy_path]['errors'][0]['message']}")

    first_spec_name = result[ksy_path]["firstSpecName"]
    first_spec = result[ksy_path]["output"]["python"][first_spec_name]
    return CompiledKSY(
        class_name=first_spec["topLevelName"],
        python_path=output_directory / first_spec["files"][0]["fileName"],
        dependencies=(
            CompiledKSY(
                class_name=compiled["topLevelName"],
                python_path=output_directory / compiled["files"][0]["fileName"]
            )
            for spec_name, compiled in result[ksy_path]["output"]["python"].items()
            if spec_name != first_spec_name
        )
    )


if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser(description="A Kaitai Struct to Python compiler")
    parser.add_argument("KSY_PATH", type=str, help="path to the Kaitai Struct definition file")
    parser.add_argument("OUTPUT_DIRECTORY", type=str, help="path to which to save the resulting Python")

    args = parser.parse_args(sys.argv[1:])

    try:
        compiled = compile(args.KSY_PATH, args.OUTPUT_DIRECTORY)
        print(f"{compiled.class_name}\t{compiled.python_path}")
        for dep in compiled.dependencies:
            print(f"{dep.class_name}\t{dep.python_path}")
    except KaitaiError as e:
        sys.stderr.write(f"{e!s}\n\n")
        sys.exit(1)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class RubyMarshal(KaitaiStruct):
    """Ruby's Marshal module allows serialization and deserialization of
    many standard and arbitrary Ruby objects in a compact binary
    format. It is relatively fast, available in stdlibs standard and
    allows conservation of language-specific properties (such as symbols
    or encoding-aware strings).
    
    Feature-wise, it is comparable to other language-specific
    implementations, such as:
    
    * Java's
      [Serializable](https://docs.oracle.com/javase/8/docs/api/java/io/Serializable.html)
    * .NET
      [BinaryFormatter](https://docs.microsoft.com/en-us/dotnet/api/system.runtime.serialization.formatters.binary.binaryformatter)
    * Python's
      [marshal](https://docs.python.org/3/library/marshal.html),
      [pickle](https://docs.python.org/3/library/pickle.html) and
      [shelve](https://docs.python.org/3/library/shelve.html)
    
    From internal perspective, serialized stream consists of a simple
    magic header and a record.
    
    .. seealso::
       Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Stream+Format
    """

    class Codes(Enum):
        ruby_string = 34
        const_nil = 48
        ruby_symbol = 58
        ruby_symbol_link = 59
        ruby_object_link = 64
        const_false = 70
        instance_var = 73
        ruby_struct = 83
        const_true = 84
        ruby_array = 91
        packed_int = 105
        bignum = 108
        ruby_hash = 123
    SEQ_FIELDS = ["version", "records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['version']['start'] = self._io.pos()
        self.version = self._io.read_bytes(2)
        self._debug['version']['end'] = self._io.pos()
        if not self.version == b"\x04\x08":
            raise kaitaistruct.ValidationNotEqualError(b"\x04\x08", self.version, self._io, u"/seq/0")
        self._debug['records']['start'] = self._io.pos()
        self.records = RubyMarshal.Record(self._io, self, self._root)
        self.records._read()
        self._debug['records']['end'] = self._io.pos()

    class RubyArray(KaitaiStruct):
        SEQ_FIELDS = ["num_elements", "elements"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_elements']['start'] = self._io.pos()
            self.num_elements = RubyMarshal.PackedInt(self._io, self, self._root)
            self.num_elements._read()
            self._debug['num_elements']['end'] = self._io.pos()
            self._debug['elements']['start'] = self._io.pos()
            self.elements = [None] * (self.num_elements.value)
            for i in range(self.num_elements.value):
                if not 'arr' in self._debug['elements']:
                    self._debug['elements']['arr'] = []
                self._debug['elements']['arr'].append({'start': self._io.pos()})
                _t_elements = RubyMarshal.Record(self._io, self, self._root)
                _t_elements._read()
                self.elements[i] = _t_elements
                self._debug['elements']['arr'][i]['end'] = self._io.pos()

            self._debug['elements']['end'] = self._io.pos()


    class Bignum(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Bignum
        """
        SEQ_FIELDS = ["sign", "len_div_2", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['sign']['start'] = self._io.pos()
            self.sign = self._io.read_u1()
            self._debug['sign']['end'] = self._io.pos()
            self._debug['len_div_2']['start'] = self._io.pos()
            self.len_div_2 = RubyMarshal.PackedInt(self._io, self, self._root)
            self.len_div_2._read()
            self._debug['len_div_2']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self.body = self._io.read_bytes((self.len_div_2.value * 2))
            self._debug['body']['end'] = self._io.pos()


    class RubyStruct(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Struct
        """
        SEQ_FIELDS = ["name", "num_members", "members"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name']['start'] = self._io.pos()
            self.name = RubyMarshal.Record(self._io, self, self._root)
            self.name._read()
            self._debug['name']['end'] = self._io.pos()
            self._debug['num_members']['start'] = self._io.pos()
            self.num_members = RubyMarshal.PackedInt(self._io, self, self._root)
            self.num_members._read()
            self._debug['num_members']['end'] = self._io.pos()
            self._debug['members']['start'] = self._io.pos()
            self.members = [None] * (self.num_members.value)
            for i in range(self.num_members.value):
                if not 'arr' in self._debug['members']:
                    self._debug['members']['arr'] = []
                self._debug['members']['arr'].append({'start': self._io.pos()})
                _t_members = RubyMarshal.Pair(self._io, self, self._root)
                _t_members._read()
                self.members[i] = _t_members
                self._debug['members']['arr'][i]['end'] = self._io.pos()

            self._debug['members']['end'] = self._io.pos()


    class RubySymbol(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Symbols+and+Byte+Sequence
        """
        SEQ_FIELDS = ["len", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = RubyMarshal.PackedInt(self._io, self, self._root)
            self.len._read()
            self._debug['len']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (self._io.read_bytes(self.len.value)).decode(u"UTF-8")
            self._debug['name']['end'] = self._io.pos()


    class PackedInt(KaitaiStruct):
        """Ruby uses sophisticated system to pack integers: first `code`
        byte either determines packing scheme or carries encoded
        immediate value (thus allowing smaller values from -123 to 122
        (inclusive) to take only one byte. There are 11 encoding schemes
        in total:
        
        * 0 is encoded specially (as 0)
        * 1..122 are encoded as immediate value with a shift
        * 123..255 are encoded with code of 0x01 and 1 extra byte
        * 0x100..0xffff are encoded with code of 0x02 and 2 extra bytes
        * 0x10000..0xffffff are encoded with code of 0x03 and 3 extra
          bytes
        * 0x1000000..0xffffffff are encoded with code of 0x04 and 4
          extra bytes
        * -123..-1 are encoded as immediate value with another shift
        * -256..-124 are encoded with code of 0xff and 1 extra byte
        * -0x10000..-257 are encoded with code of 0xfe and 2 extra bytes
        * -0x1000000..0x10001 are encoded with code of 0xfd and 3 extra
           bytes
        * -0x40000000..-0x1000001 are encoded with code of 0xfc and 4
           extra bytes
        
        Values beyond that are serialized as bignum (even if they
        technically might be not Bignum class in Ruby implementation,
        i.e. if they fit into 64 bits on a 64-bit platform).
        
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Fixnum+and+long
        """
        SEQ_FIELDS = ["code", "encoded", "encoded2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = self._io.read_u1()
            self._debug['code']['end'] = self._io.pos()
            self._debug['encoded']['start'] = self._io.pos()
            _on = self.code
            if _on == 4:
                self.encoded = self._io.read_u4le()
            elif _on == 1:
                self.encoded = self._io.read_u1()
            elif _on == 252:
                self.encoded = self._io.read_u4le()
            elif _on == 253:
                self.encoded = self._io.read_u2le()
            elif _on == 3:
                self.encoded = self._io.read_u2le()
            elif _on == 2:
                self.encoded = self._io.read_u2le()
            elif _on == 255:
                self.encoded = self._io.read_u1()
            elif _on == 254:
                self.encoded = self._io.read_u2le()
            self._debug['encoded']['end'] = self._io.pos()
            self._debug['encoded2']['start'] = self._io.pos()
            _on = self.code
            if _on == 3:
                self.encoded2 = self._io.read_u1()
            elif _on == 253:
                self.encoded2 = self._io.read_u1()
            self._debug['encoded2']['end'] = self._io.pos()

        @property
        def is_immediate(self):
            if hasattr(self, '_m_is_immediate'):
                return self._m_is_immediate if hasattr(self, '_m_is_immediate') else None

            self._m_is_immediate =  ((self.code > 4) and (self.code < 252)) 
            return self._m_is_immediate if hasattr(self, '_m_is_immediate') else None

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = (((self.code - 5) if self.code < 128 else (4 - (~(self.code) & 127))) if self.is_immediate else (0 if self.code == 0 else ((self.encoded - 256) if self.code == 255 else ((self.encoded - 65536) if self.code == 254 else ((((self.encoded2 << 16) | self.encoded) - 16777216) if self.code == 253 else (((self.encoded2 << 16) | self.encoded) if self.code == 3 else self.encoded))))))
            return self._m_value if hasattr(self, '_m_value') else None


    class Pair(KaitaiStruct):
        SEQ_FIELDS = ["key", "value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['key']['start'] = self._io.pos()
            self.key = RubyMarshal.Record(self._io, self, self._root)
            self.key._read()
            self._debug['key']['end'] = self._io.pos()
            self._debug['value']['start'] = self._io.pos()
            self.value = RubyMarshal.Record(self._io, self, self._root)
            self.value._read()
            self._debug['value']['end'] = self._io.pos()


    class InstanceVar(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Instance+Variables
        """
        SEQ_FIELDS = ["obj", "num_vars", "vars"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['obj']['start'] = self._io.pos()
            self.obj = RubyMarshal.Record(self._io, self, self._root)
            self.obj._read()
            self._debug['obj']['end'] = self._io.pos()
            self._debug['num_vars']['start'] = self._io.pos()
            self.num_vars = RubyMarshal.PackedInt(self._io, self, self._root)
            self.num_vars._read()
            self._debug['num_vars']['end'] = self._io.pos()
            self._debug['vars']['start'] = self._io.pos()
            self.vars = [None] * (self.num_vars.value)
            for i in range(self.num_vars.value):
                if not 'arr' in self._debug['vars']:
                    self._debug['vars']['arr'] = []
                self._debug['vars']['arr'].append({'start': self._io.pos()})
                _t_vars = RubyMarshal.Pair(self._io, self, self._root)
                _t_vars._read()
                self.vars[i] = _t_vars
                self._debug['vars']['arr'][i]['end'] = self._io.pos()

            self._debug['vars']['end'] = self._io.pos()


    class Record(KaitaiStruct):
        """Each record starts with a single byte that determines its type
        (`code`) and contents. If necessary, additional info as parsed
        as `body`, to be determined by `code`.
        """
        SEQ_FIELDS = ["code", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = KaitaiStream.resolve_enum(RubyMarshal.Codes, self._io.read_u1())
            self._debug['code']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.code
            if _on == RubyMarshal.Codes.packed_int:
                self.body = RubyMarshal.PackedInt(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.bignum:
                self.body = RubyMarshal.Bignum(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_array:
                self.body = RubyMarshal.RubyArray(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_symbol_link:
                self.body = RubyMarshal.PackedInt(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_struct:
                self.body = RubyMarshal.RubyStruct(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_string:
                self.body = RubyMarshal.RubyString(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.instance_var:
                self.body = RubyMarshal.InstanceVar(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_hash:
                self.body = RubyMarshal.RubyHash(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_symbol:
                self.body = RubyMarshal.RubySymbol(self._io, self, self._root)
                self.body._read()
            elif _on == RubyMarshal.Codes.ruby_object_link:
                self.body = RubyMarshal.PackedInt(self._io, self, self._root)
                self.body._read()
            self._debug['body']['end'] = self._io.pos()


    class RubyHash(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-Hash+and+Hash+with+Default+Value
        """
        SEQ_FIELDS = ["num_pairs", "pairs"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_pairs']['start'] = self._io.pos()
            self.num_pairs = RubyMarshal.PackedInt(self._io, self, self._root)
            self.num_pairs._read()
            self._debug['num_pairs']['end'] = self._io.pos()
            self._debug['pairs']['start'] = self._io.pos()
            self.pairs = [None] * (self.num_pairs.value)
            for i in range(self.num_pairs.value):
                if not 'arr' in self._debug['pairs']:
                    self._debug['pairs']['arr'] = []
                self._debug['pairs']['arr'].append({'start': self._io.pos()})
                _t_pairs = RubyMarshal.Pair(self._io, self, self._root)
                _t_pairs._read()
                self.pairs[i] = _t_pairs
                self._debug['pairs']['arr'][i]['end'] = self._io.pos()

            self._debug['pairs']['end'] = self._io.pos()


    class RubyString(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.ruby-lang.org/en/2.4.0/marshal_rdoc.html#label-String
        """
        SEQ_FIELDS = ["len", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = RubyMarshal.PackedInt(self._io, self, self._root)
            self.len._read()
            self._debug['len']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self.body = self._io.read_bytes(self.len.value)
            self._debug['body']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class QuakeMdl(KaitaiStruct):
    SEQ_FIELDS = ["header", "skins", "texture_coordinates", "triangles", "frames"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = QuakeMdl.MdlHeader(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['skins']['start'] = self._io.pos()
        self.skins = [None] * (self.header.num_skins)
        for i in range(self.header.num_skins):
            if not 'arr' in self._debug['skins']:
                self._debug['skins']['arr'] = []
            self._debug['skins']['arr'].append({'start': self._io.pos()})
            _t_skins = QuakeMdl.MdlSkin(self._io, self, self._root)
            _t_skins._read()
            self.skins[i] = _t_skins
            self._debug['skins']['arr'][i]['end'] = self._io.pos()

        self._debug['skins']['end'] = self._io.pos()
        self._debug['texture_coordinates']['start'] = self._io.pos()
        self.texture_coordinates = [None] * (self.header.num_verts)
        for i in range(self.header.num_verts):
            if not 'arr' in self._debug['texture_coordinates']:
                self._debug['texture_coordinates']['arr'] = []
            self._debug['texture_coordinates']['arr'].append({'start': self._io.pos()})
            _t_texture_coordinates = QuakeMdl.MdlTexcoord(self._io, self, self._root)
            _t_texture_coordinates._read()
            self.texture_coordinates[i] = _t_texture_coordinates
            self._debug['texture_coordinates']['arr'][i]['end'] = self._io.pos()

        self._debug['texture_coordinates']['end'] = self._io.pos()
        self._debug['triangles']['start'] = self._io.pos()
        self.triangles = [None] * (self.header.num_tris)
        for i in range(self.header.num_tris):
            if not 'arr' in self._debug['triangles']:
                self._debug['triangles']['arr'] = []
            self._debug['triangles']['arr'].append({'start': self._io.pos()})
            _t_triangles = QuakeMdl.MdlTriangle(self._io, self, self._root)
            _t_triangles._read()
            self.triangles[i] = _t_triangles
            self._debug['triangles']['arr'][i]['end'] = self._io.pos()

        self._debug['triangles']['end'] = self._io.pos()
        self._debug['frames']['start'] = self._io.pos()
        self.frames = [None] * (self.header.num_frames)
        for i in range(self.header.num_frames):
            if not 'arr' in self._debug['frames']:
                self._debug['frames']['arr'] = []
            self._debug['frames']['arr'].append({'start': self._io.pos()})
            _t_frames = QuakeMdl.MdlFrame(self._io, self, self._root)
            _t_frames._read()
            self.frames[i] = _t_frames
            self._debug['frames']['arr'][i]['end'] = self._io.pos()

        self._debug['frames']['end'] = self._io.pos()

    class MdlVertex(KaitaiStruct):
        SEQ_FIELDS = ["values", "normal_index"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['values']['start'] = self._io.pos()
            self.values = [None] * (3)
            for i in range(3):
                if not 'arr' in self._debug['values']:
                    self._debug['values']['arr'] = []
                self._debug['values']['arr'].append({'start': self._io.pos()})
                self.values[i] = self._io.read_u1()
                self._debug['values']['arr'][i]['end'] = self._io.pos()

            self._debug['values']['end'] = self._io.pos()
            self._debug['normal_index']['start'] = self._io.pos()
            self.normal_index = self._io.read_u1()
            self._debug['normal_index']['end'] = self._io.pos()


    class MdlTexcoord(KaitaiStruct):
        SEQ_FIELDS = ["on_seam", "s", "t"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['on_seam']['start'] = self._io.pos()
            self.on_seam = self._io.read_s4le()
            self._debug['on_seam']['end'] = self._io.pos()
            self._debug['s']['start'] = self._io.pos()
            self.s = self._io.read_s4le()
            self._debug['s']['end'] = self._io.pos()
            self._debug['t']['start'] = self._io.pos()
            self.t = self._io.read_s4le()
            self._debug['t']['end'] = self._io.pos()


    class MdlHeader(KaitaiStruct):
        SEQ_FIELDS = ["ident", "version_must_be_6", "scale", "origin", "radius", "eye_position", "num_skins", "skin_width", "skin_height", "num_verts", "num_tris", "num_frames", "synctype", "flags", "size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ident']['start'] = self._io.pos()
            self.ident = self._io.read_bytes(4)
            self._debug['ident']['end'] = self._io.pos()
            if not self.ident == b"\x49\x44\x50\x4F":
                raise kaitaistruct.ValidationNotEqualError(b"\x49\x44\x50\x4F", self.ident, self._io, u"/types/mdl_header/seq/0")
            self._debug['version_must_be_6']['start'] = self._io.pos()
            self.version_must_be_6 = self._io.read_bytes(4)
            self._debug['version_must_be_6']['end'] = self._io.pos()
            if not self.version_must_be_6 == b"\x06\x00\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x06\x00\x00\x00", self.version_must_be_6, self._io, u"/types/mdl_header/seq/1")
            self._debug['scale']['start'] = self._io.pos()
            self.scale = QuakeMdl.Vec3(self._io, self, self._root)
            self.scale._read()
            self._debug['scale']['end'] = self._io.pos()
            self._debug['origin']['start'] = self._io.pos()
            self.origin = QuakeMdl.Vec3(self._io, self, self._root)
            self.origin._read()
            self._debug['origin']['end'] = self._io.pos()
            self._debug['radius']['start'] = self._io.pos()
            self.radius = self._io.read_f4le()
            self._debug['radius']['end'] = self._io.pos()
            self._debug['eye_position']['start'] = self._io.pos()
            self.eye_position = QuakeMdl.Vec3(self._io, self, self._root)
            self.eye_position._read()
            self._debug['eye_position']['end'] = self._io.pos()
            self._debug['num_skins']['start'] = self._io.pos()
            self.num_skins = self._io.read_s4le()
            self._debug['num_skins']['end'] = self._io.pos()
            self._debug['skin_width']['start'] = self._io.pos()
            self.skin_width = self._io.read_s4le()
            self._debug['skin_width']['end'] = self._io.pos()
            self._debug['skin_height']['start'] = self._io.pos()
            self.skin_height = self._io.read_s4le()
            self._debug['skin_height']['end'] = self._io.pos()
            self._debug['num_verts']['start'] = self._io.pos()
            self.num_verts = self._io.read_s4le()
            self._debug['num_verts']['end'] = self._io.pos()
            self._debug['num_tris']['start'] = self._io.pos()
            self.num_tris = self._io.read_s4le()
            self._debug['num_tris']['end'] = self._io.pos()
            self._debug['num_frames']['start'] = self._io.pos()
            self.num_frames = self._io.read_s4le()
            self._debug['num_frames']['end'] = self._io.pos()
            self._debug['synctype']['start'] = self._io.pos()
            self.synctype = self._io.read_s4le()
            self._debug['synctype']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_s4le()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_f4le()
            self._debug['size']['end'] = self._io.pos()

        @property
        def version(self):
            if hasattr(self, '_m_version'):
                return self._m_version if hasattr(self, '_m_version') else None

            self._m_version = 6
            return self._m_version if hasattr(self, '_m_version') else None

        @property
        def skin_size(self):
            if hasattr(self, '_m_skin_size'):
                return self._m_skin_size if hasattr(self, '_m_skin_size') else None

            self._m_skin_size = (self.skin_width * self.skin_height)
            return self._m_skin_size if hasattr(self, '_m_skin_size') else None


    class MdlSkin(KaitaiStruct):
        SEQ_FIELDS = ["group", "single_texture_data", "num_frames", "frame_times", "group_texture_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['group']['start'] = self._io.pos()
            self.group = self._io.read_s4le()
            self._debug['group']['end'] = self._io.pos()
            if self.group == 0:
                self._debug['single_texture_data']['start'] = self._io.pos()
                self.single_texture_data = self._io.read_bytes(self._root.header.skin_size)
                self._debug['single_texture_data']['end'] = self._io.pos()

            if self.group != 0:
                self._debug['num_frames']['start'] = self._io.pos()
                self.num_frames = self._io.read_u4le()
                self._debug['num_frames']['end'] = self._io.pos()

            if self.group != 0:
                self._debug['frame_times']['start'] = self._io.pos()
                self.frame_times = [None] * (self.num_frames)
                for i in range(self.num_frames):
                    if not 'arr' in self._debug['frame_times']:
                        self._debug['frame_times']['arr'] = []
                    self._debug['frame_times']['arr'].append({'start': self._io.pos()})
                    self.frame_times[i] = self._io.read_f4le()
                    self._debug['frame_times']['arr'][i]['end'] = self._io.pos()

                self._debug['frame_times']['end'] = self._io.pos()

            if self.group != 0:
                self._debug['group_texture_data']['start'] = self._io.pos()
                self.group_texture_data = [None] * (self.num_frames)
                for i in range(self.num_frames):
                    if not 'arr' in self._debug['group_texture_data']:
                        self._debug['group_texture_data']['arr'] = []
                    self._debug['group_texture_data']['arr'].append({'start': self._io.pos()})
                    self.group_texture_data[i] = self._io.read_bytes(self._root.header.skin_size)
                    self._debug['group_texture_data']['arr'][i]['end'] = self._io.pos()

                self._debug['group_texture_data']['end'] = self._io.pos()



    class MdlFrame(KaitaiStruct):
        SEQ_FIELDS = ["type", "min", "max", "time", "frames"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = self._io.read_s4le()
            self._debug['type']['end'] = self._io.pos()
            if self.type != 0:
                self._debug['min']['start'] = self._io.pos()
                self.min = QuakeMdl.MdlVertex(self._io, self, self._root)
                self.min._read()
                self._debug['min']['end'] = self._io.pos()

            if self.type != 0:
                self._debug['max']['start'] = self._io.pos()
                self.max = QuakeMdl.MdlVertex(self._io, self, self._root)
                self.max._read()
                self._debug['max']['end'] = self._io.pos()

            if self.type != 0:
                self._debug['time']['start'] = self._io.pos()
                self.time = [None] * (self.type)
                for i in range(self.type):
                    if not 'arr' in self._debug['time']:
                        self._debug['time']['arr'] = []
                    self._debug['time']['arr'].append({'start': self._io.pos()})
                    self.time[i] = self._io.read_f4le()
                    self._debug['time']['arr'][i]['end'] = self._io.pos()

                self._debug['time']['end'] = self._io.pos()

            self._debug['frames']['start'] = self._io.pos()
            self.frames = [None] * (self.num_simple_frames)
            for i in range(self.num_simple_frames):
                if not 'arr' in self._debug['frames']:
                    self._debug['frames']['arr'] = []
                self._debug['frames']['arr'].append({'start': self._io.pos()})
                _t_frames = QuakeMdl.MdlSimpleFrame(self._io, self, self._root)
                _t_frames._read()
                self.frames[i] = _t_frames
                self._debug['frames']['arr'][i]['end'] = self._io.pos()

            self._debug['frames']['end'] = self._io.pos()

        @property
        def num_simple_frames(self):
            if hasattr(self, '_m_num_simple_frames'):
                return self._m_num_simple_frames if hasattr(self, '_m_num_simple_frames') else None

            self._m_num_simple_frames = (1 if self.type == 0 else self.type)
            return self._m_num_simple_frames if hasattr(self, '_m_num_simple_frames') else None


    class MdlSimpleFrame(KaitaiStruct):
        SEQ_FIELDS = ["bbox_min", "bbox_max", "name", "vertices"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['bbox_min']['start'] = self._io.pos()
            self.bbox_min = QuakeMdl.MdlVertex(self._io, self, self._root)
            self.bbox_min._read()
            self._debug['bbox_min']['end'] = self._io.pos()
            self._debug['bbox_max']['start'] = self._io.pos()
            self.bbox_max = QuakeMdl.MdlVertex(self._io, self, self._root)
            self.bbox_max._read()
            self._debug['bbox_max']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (KaitaiStream.bytes_terminate(KaitaiStream.bytes_strip_right(self._io.read_bytes(16), 0), 0, False)).decode(u"ASCII")
            self._debug['name']['end'] = self._io.pos()
            self._debug['vertices']['start'] = self._io.pos()
            self.vertices = [None] * (self._root.header.num_verts)
            for i in range(self._root.header.num_verts):
                if not 'arr' in self._debug['vertices']:
                    self._debug['vertices']['arr'] = []
                self._debug['vertices']['arr'].append({'start': self._io.pos()})
                _t_vertices = QuakeMdl.MdlVertex(self._io, self, self._root)
                _t_vertices._read()
                self.vertices[i] = _t_vertices
                self._debug['vertices']['arr'][i]['end'] = self._io.pos()

            self._debug['vertices']['end'] = self._io.pos()


    class MdlTriangle(KaitaiStruct):
        SEQ_FIELDS = ["faces_front", "vertices"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['faces_front']['start'] = self._io.pos()
            self.faces_front = self._io.read_s4le()
            self._debug['faces_front']['end'] = self._io.pos()
            self._debug['vertices']['start'] = self._io.pos()
            self.vertices = [None] * (3)
            for i in range(3):
                if not 'arr' in self._debug['vertices']:
                    self._debug['vertices']['arr'] = []
                self._debug['vertices']['arr'].append({'start': self._io.pos()})
                self.vertices[i] = self._io.read_s4le()
                self._debug['vertices']['arr'][i]['end'] = self._io.pos()

            self._debug['vertices']['end'] = self._io.pos()


    class Vec3(KaitaiStruct):
        SEQ_FIELDS = ["x", "y", "z"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_f4le()
            self._debug['x']['end'] = self._io.pos()
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_f4le()
            self._debug['y']['end'] = self._io.pos()
            self._debug['z']['start'] = self._io.pos()
            self.z = self._io.read_f4le()
            self._debug['z']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class NtMdt(KaitaiStruct):
    """A native file format of NT-MDT scientific software. Usually contains
    any of:
    
    * [Scanning probe](https://en.wikipedia.org/wiki/Scanning_probe_microscopy) microscopy scans and spectra
    * [Raman spectra](https://en.wikipedia.org/wiki/Raman_spectroscopy)
    * results of their analysis
    
    Some examples of mdt files can be downloaded at:
    
    * http://www.ntmdt-si.ru/scan-gallery
    * http://callistosoft.narod.ru/Resources/Mdt.zip
    
    .. seealso::
       Source - https://svn.code.sf.net/p/gwyddion/code/trunk/gwyddion/modules/file/nt-mdt.c
    """

    class AdcMode(Enum):
        height = 0
        dfl = 1
        lateral_f = 2
        bias_v = 3
        current = 4
        fb_out = 5
        mag = 6
        mag_sin = 7
        mag_cos = 8
        rms = 9
        calc_mag = 10
        phase1 = 11
        phase2 = 12
        calc_phase = 13
        ex1 = 14
        ex2 = 15
        hv_x = 16
        hv_y = 17
        snap_back = 18
        false = 255

    class XmlScanLocation(Enum):
        hlt = 0
        hlb = 1
        hrt = 2
        hrb = 3
        vlt = 4
        vlb = 5
        vrt = 6
        vrb = 7

    class DataType(Enum):
        floatfix = -65544
        float80 = -16138
        float64 = -13320
        float48 = -9990
        float32 = -5892
        int64 = -8
        int32 = -4
        int16 = -2
        int8 = -1
        unknown0 = 0
        uint8 = 1
        uint16 = 2
        uint32 = 4
        uint64 = 8

    class XmlParamType(Enum):
        none = 0
        laser_wavelength = 1
        units = 2
        data_array = 255

    class SpmMode(Enum):
        constant_force = 0
        contact_constant_height = 1
        contact_error = 2
        lateral_force = 3
        force_modulation = 4
        spreading_resistance_imaging = 5
        semicontact_topography = 6
        semicontact_error = 7
        phase_contrast = 8
        ac_magnetic_force = 9
        dc_magnetic_force = 10
        electrostatic_force = 11
        capacitance_contrast = 12
        kelvin_probe = 13
        constant_current = 14
        barrier_height = 15
        constant_height = 16
        afam = 17
        contact_efm = 18
        shear_force_topography = 19
        sfom = 20
        contact_capacitance = 21
        snom_transmission = 22
        snom_reflection = 23
        snom_all = 24
        snom = 25

    class Unit(Enum):
        raman_shift = -10
        reserved0 = -9
        reserved1 = -8
        reserved2 = -7
        reserved3 = -6
        meter = -5
        centi_meter = -4
        milli_meter = -3
        micro_meter = -2
        nano_meter = -1
        angstrom = 0
        nano_ampere = 1
        volt = 2
        none = 3
        kilo_hertz = 4
        degrees = 5
        percent = 6
        celsius_degree = 7
        volt_high = 8
        second = 9
        milli_second = 10
        micro_second = 11
        nano_second = 12
        counts = 13
        pixels = 14
        reserved_sfom0 = 15
        reserved_sfom1 = 16
        reserved_sfom2 = 17
        reserved_sfom3 = 18
        reserved_sfom4 = 19
        ampere2 = 20
        milli_ampere = 21
        micro_ampere = 22
        nano_ampere2 = 23
        pico_ampere = 24
        volt2 = 25
        milli_volt = 26
        micro_volt = 27
        nano_volt = 28
        pico_volt = 29
        newton = 30
        milli_newton = 31
        micro_newton = 32
        nano_newton = 33
        pico_newton = 34
        reserved_dos0 = 35
        reserved_dos1 = 36
        reserved_dos2 = 37
        reserved_dos3 = 38
        reserved_dos4 = 39

    class SpmTechnique(Enum):
        contact_mode = 0
        semicontact_mode = 1
        tunnel_current = 2
        snom = 3

    class Consts(Enum):
        frame_mode_size = 8
        frame_header_size = 22
        axis_scales_size = 30
        file_header_size = 32
        spectro_vars_min_size = 38
        scan_vars_min_size = 77
    SEQ_FIELDS = ["signature", "size", "reserved0", "last_frame", "reserved1", "wrond_doc", "frames"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['signature']['start'] = self._io.pos()
        self.signature = self._io.read_bytes(4)
        self._debug['signature']['end'] = self._io.pos()
        if not self.signature == b"\x01\xB0\x93\xFF":
            raise kaitaistruct.ValidationNotEqualError(b"\x01\xB0\x93\xFF", self.signature, self._io, u"/seq/0")
        self._debug['size']['start'] = self._io.pos()
        self.size = self._io.read_u4le()
        self._debug['size']['end'] = self._io.pos()
        self._debug['reserved0']['start'] = self._io.pos()
        self.reserved0 = self._io.read_bytes(4)
        self._debug['reserved0']['end'] = self._io.pos()
        self._debug['last_frame']['start'] = self._io.pos()
        self.last_frame = self._io.read_u2le()
        self._debug['last_frame']['end'] = self._io.pos()
        self._debug['reserved1']['start'] = self._io.pos()
        self.reserved1 = self._io.read_bytes(18)
        self._debug['reserved1']['end'] = self._io.pos()
        self._debug['wrond_doc']['start'] = self._io.pos()
        self.wrond_doc = self._io.read_bytes(1)
        self._debug['wrond_doc']['end'] = self._io.pos()
        self._debug['frames']['start'] = self._io.pos()
        self._raw_frames = self._io.read_bytes(self.size)
        _io__raw_frames = KaitaiStream(BytesIO(self._raw_frames))
        self.frames = NtMdt.Framez(_io__raw_frames, self, self._root)
        self.frames._read()
        self._debug['frames']['end'] = self._io.pos()

    class Uuid(KaitaiStruct):
        SEQ_FIELDS = ["data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['data']['start'] = self._io.pos()
            self.data = [None] * (16)
            for i in range(16):
                if not 'arr' in self._debug['data']:
                    self._debug['data']['arr'] = []
                self._debug['data']['arr'].append({'start': self._io.pos()})
                self.data[i] = self._io.read_u1()
                self._debug['data']['arr'][i]['end'] = self._io.pos()

            self._debug['data']['end'] = self._io.pos()


    class Framez(KaitaiStruct):
        SEQ_FIELDS = ["frames"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['frames']['start'] = self._io.pos()
            self.frames = [None] * ((self._root.last_frame + 1))
            for i in range((self._root.last_frame + 1)):
                if not 'arr' in self._debug['frames']:
                    self._debug['frames']['arr'] = []
                self._debug['frames']['arr'].append({'start': self._io.pos()})
                _t_frames = NtMdt.Frame(self._io, self, self._root)
                _t_frames._read()
                self.frames[i] = _t_frames
                self._debug['frames']['arr'][i]['end'] = self._io.pos()

            self._debug['frames']['end'] = self._io.pos()


    class Frame(KaitaiStruct):

        class FrameType(Enum):
            scanned = 0
            spectroscopy = 1
            text = 3
            old_mda = 105
            mda = 106
            palette = 107
            curves_new = 190
            curves = 201
        SEQ_FIELDS = ["size", "main"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['main']['start'] = self._io.pos()
            self._raw_main = self._io.read_bytes((self.size - 4))
            _io__raw_main = KaitaiStream(BytesIO(self._raw_main))
            self.main = NtMdt.Frame.FrameMain(_io__raw_main, self, self._root)
            self.main._read()
            self._debug['main']['end'] = self._io.pos()

        class Dots(KaitaiStruct):
            SEQ_FIELDS = ["fm_ndots", "coord_header", "coordinates", "data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['fm_ndots']['start'] = self._io.pos()
                self.fm_ndots = self._io.read_u2le()
                self._debug['fm_ndots']['end'] = self._io.pos()
                if self.fm_ndots > 0:
                    self._debug['coord_header']['start'] = self._io.pos()
                    self.coord_header = NtMdt.Frame.Dots.DotsHeader(self._io, self, self._root)
                    self.coord_header._read()
                    self._debug['coord_header']['end'] = self._io.pos()

                self._debug['coordinates']['start'] = self._io.pos()
                self.coordinates = [None] * (self.fm_ndots)
                for i in range(self.fm_ndots):
                    if not 'arr' in self._debug['coordinates']:
                        self._debug['coordinates']['arr'] = []
                    self._debug['coordinates']['arr'].append({'start': self._io.pos()})
                    _t_coordinates = NtMdt.Frame.Dots.DotsData(self._io, self, self._root)
                    _t_coordinates._read()
                    self.coordinates[i] = _t_coordinates
                    self._debug['coordinates']['arr'][i]['end'] = self._io.pos()

                self._debug['coordinates']['end'] = self._io.pos()
                self._debug['data']['start'] = self._io.pos()
                self.data = [None] * (self.fm_ndots)
                for i in range(self.fm_ndots):
                    if not 'arr' in self._debug['data']:
                        self._debug['data']['arr'] = []
                    self._debug['data']['arr'].append({'start': self._io.pos()})
                    _t_data = NtMdt.Frame.Dots.DataLinez(i, self._io, self, self._root)
                    _t_data._read()
                    self.data[i] = _t_data
                    self._debug['data']['arr'][i]['end'] = self._io.pos()

                self._debug['data']['end'] = self._io.pos()

            class DotsHeader(KaitaiStruct):
                SEQ_FIELDS = ["header_size", "header"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['header_size']['start'] = self._io.pos()
                    self.header_size = self._io.read_s4le()
                    self._debug['header_size']['end'] = self._io.pos()
                    self._debug['header']['start'] = self._io.pos()
                    self._raw_header = self._io.read_bytes(self.header_size)
                    _io__raw_header = KaitaiStream(BytesIO(self._raw_header))
                    self.header = NtMdt.Frame.Dots.DotsHeader.Header(_io__raw_header, self, self._root)
                    self.header._read()
                    self._debug['header']['end'] = self._io.pos()

                class Header(KaitaiStruct):
                    SEQ_FIELDS = ["coord_size", "version", "xyunits"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['coord_size']['start'] = self._io.pos()
                        self.coord_size = self._io.read_s4le()
                        self._debug['coord_size']['end'] = self._io.pos()
                        self._debug['version']['start'] = self._io.pos()
                        self.version = self._io.read_s4le()
                        self._debug['version']['end'] = self._io.pos()
                        self._debug['xyunits']['start'] = self._io.pos()
                        self.xyunits = KaitaiStream.resolve_enum(NtMdt.Unit, self._io.read_s2le())
                        self._debug['xyunits']['end'] = self._io.pos()



            class DotsData(KaitaiStruct):
                SEQ_FIELDS = ["coord_x", "coord_y", "forward_size", "backward_size"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['coord_x']['start'] = self._io.pos()
                    self.coord_x = self._io.read_f4le()
                    self._debug['coord_x']['end'] = self._io.pos()
                    self._debug['coord_y']['start'] = self._io.pos()
                    self.coord_y = self._io.read_f4le()
                    self._debug['coord_y']['end'] = self._io.pos()
                    self._debug['forward_size']['start'] = self._io.pos()
                    self.forward_size = self._io.read_s4le()
                    self._debug['forward_size']['end'] = self._io.pos()
                    self._debug['backward_size']['start'] = self._io.pos()
                    self.backward_size = self._io.read_s4le()
                    self._debug['backward_size']['end'] = self._io.pos()


            class DataLinez(KaitaiStruct):
                SEQ_FIELDS = ["forward", "backward"]
                def __init__(self, index, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self.index = index
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['forward']['start'] = self._io.pos()
                    self.forward = [None] * (self._parent.coordinates[self.index].forward_size)
                    for i in range(self._parent.coordinates[self.index].forward_size):
                        if not 'arr' in self._debug['forward']:
                            self._debug['forward']['arr'] = []
                        self._debug['forward']['arr'].append({'start': self._io.pos()})
                        self.forward[i] = self._io.read_s2le()
                        self._debug['forward']['arr'][i]['end'] = self._io.pos()

                    self._debug['forward']['end'] = self._io.pos()
                    self._debug['backward']['start'] = self._io.pos()
                    self.backward = [None] * (self._parent.coordinates[self.index].backward_size)
                    for i in range(self._parent.coordinates[self.index].backward_size):
                        if not 'arr' in self._debug['backward']:
                            self._debug['backward']['arr'] = []
                        self._debug['backward']['arr'].append({'start': self._io.pos()})
                        self.backward[i] = self._io.read_s2le()
                        self._debug['backward']['arr'][i]['end'] = self._io.pos()

                    self._debug['backward']['end'] = self._io.pos()



        class FrameMain(KaitaiStruct):
            SEQ_FIELDS = ["type", "version", "date_time", "var_size", "frame_data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['type']['start'] = self._io.pos()
                self.type = KaitaiStream.resolve_enum(NtMdt.Frame.FrameType, self._io.read_u2le())
                self._debug['type']['end'] = self._io.pos()
                self._debug['version']['start'] = self._io.pos()
                self.version = NtMdt.Version(self._io, self, self._root)
                self.version._read()
                self._debug['version']['end'] = self._io.pos()
                self._debug['date_time']['start'] = self._io.pos()
                self.date_time = NtMdt.Frame.DateTime(self._io, self, self._root)
                self.date_time._read()
                self._debug['date_time']['end'] = self._io.pos()
                self._debug['var_size']['start'] = self._io.pos()
                self.var_size = self._io.read_u2le()
                self._debug['var_size']['end'] = self._io.pos()
                self._debug['frame_data']['start'] = self._io.pos()
                _on = self.type
                if _on == NtMdt.Frame.FrameType.mda:
                    self._raw_frame_data = self._io.read_bytes_full()
                    _io__raw_frame_data = KaitaiStream(BytesIO(self._raw_frame_data))
                    self.frame_data = NtMdt.Frame.FdMetaData(_io__raw_frame_data, self, self._root)
                    self.frame_data._read()
                elif _on == NtMdt.Frame.FrameType.curves_new:
                    self._raw_frame_data = self._io.read_bytes_full()
                    _io__raw_frame_data = KaitaiStream(BytesIO(self._raw_frame_data))
                    self.frame_data = NtMdt.Frame.FdCurvesNew(_io__raw_frame_data, self, self._root)
                    self.frame_data._read()
                elif _on == NtMdt.Frame.FrameType.curves:
                    self._raw_frame_data = self._io.read_bytes_full()
                    _io__raw_frame_data = KaitaiStream(BytesIO(self._raw_frame_data))
                    self.frame_data = NtMdt.Frame.FdSpectroscopy(_io__raw_frame_data, self, self._root)
                    self.frame_data._read()
                elif _on == NtMdt.Frame.FrameType.spectroscopy:
                    self._raw_frame_data = self._io.read_bytes_full()
                    _io__raw_frame_data = KaitaiStream(BytesIO(self._raw_frame_data))
                    self.frame_data = NtMdt.Frame.FdSpectroscopy(_io__raw_frame_data, self, self._root)
                    self.frame_data._read()
                elif _on == NtMdt.Frame.FrameType.scanned:
                    self._raw_frame_data = self._io.read_bytes_full()
                    _io__raw_frame_data = KaitaiStream(BytesIO(self._raw_frame_data))
                    self.frame_data = NtMdt.Frame.FdScanned(_io__raw_frame_data, self, self._root)
                    self.frame_data._read()
                else:
                    self.frame_data = self._io.read_bytes_full()
                self._debug['frame_data']['end'] = self._io.pos()


        class FdCurvesNew(KaitaiStruct):
            SEQ_FIELDS = ["block_count", "blocks_headers", "blocks_names", "blocks_data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['block_count']['start'] = self._io.pos()
                self.block_count = self._io.read_u4le()
                self._debug['block_count']['end'] = self._io.pos()
                self._debug['blocks_headers']['start'] = self._io.pos()
                self.blocks_headers = [None] * (self.block_count)
                for i in range(self.block_count):
                    if not 'arr' in self._debug['blocks_headers']:
                        self._debug['blocks_headers']['arr'] = []
                    self._debug['blocks_headers']['arr'].append({'start': self._io.pos()})
                    _t_blocks_headers = NtMdt.Frame.FdCurvesNew.BlockDescr(self._io, self, self._root)
                    _t_blocks_headers._read()
                    self.blocks_headers[i] = _t_blocks_headers
                    self._debug['blocks_headers']['arr'][i]['end'] = self._io.pos()

                self._debug['blocks_headers']['end'] = self._io.pos()
                self._debug['blocks_names']['start'] = self._io.pos()
                self.blocks_names = [None] * (self.block_count)
                for i in range(self.block_count):
                    if not 'arr' in self._debug['blocks_names']:
                        self._debug['blocks_names']['arr'] = []
                    self._debug['blocks_names']['arr'].append({'start': self._io.pos()})
                    self.blocks_names[i] = (self._io.read_bytes(self.blocks_headers[i].name_len)).decode(u"UTF-8")
                    self._debug['blocks_names']['arr'][i]['end'] = self._io.pos()

                self._debug['blocks_names']['end'] = self._io.pos()
                self._debug['blocks_data']['start'] = self._io.pos()
                self.blocks_data = [None] * (self.block_count)
                for i in range(self.block_count):
                    if not 'arr' in self._debug['blocks_data']:
                        self._debug['blocks_data']['arr'] = []
                    self._debug['blocks_data']['arr'].append({'start': self._io.pos()})
                    self.blocks_data[i] = self._io.read_bytes(self.blocks_headers[i].len)
                    self._debug['blocks_data']['arr'][i]['end'] = self._io.pos()

                self._debug['blocks_data']['end'] = self._io.pos()

            class BlockDescr(KaitaiStruct):
                SEQ_FIELDS = ["name_len", "len"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['name_len']['start'] = self._io.pos()
                    self.name_len = self._io.read_u4le()
                    self._debug['name_len']['end'] = self._io.pos()
                    self._debug['len']['start'] = self._io.pos()
                    self.len = self._io.read_u4le()
                    self._debug['len']['end'] = self._io.pos()



        class FdMetaData(KaitaiStruct):
            SEQ_FIELDS = ["head_size", "tot_len", "guids", "frame_status", "name_size", "comm_size", "view_info_size", "spec_size", "source_info_size", "var_size", "data_offset", "data_size", "title", "xml", "struct_len", "array_size", "cell_size", "n_dimensions", "n_mesurands", "dimensions", "mesurands"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['head_size']['start'] = self._io.pos()
                self.head_size = self._io.read_u4le()
                self._debug['head_size']['end'] = self._io.pos()
                self._debug['tot_len']['start'] = self._io.pos()
                self.tot_len = self._io.read_u4le()
                self._debug['tot_len']['end'] = self._io.pos()
                self._debug['guids']['start'] = self._io.pos()
                self.guids = [None] * (2)
                for i in range(2):
                    if not 'arr' in self._debug['guids']:
                        self._debug['guids']['arr'] = []
                    self._debug['guids']['arr'].append({'start': self._io.pos()})
                    _t_guids = NtMdt.Uuid(self._io, self, self._root)
                    _t_guids._read()
                    self.guids[i] = _t_guids
                    self._debug['guids']['arr'][i]['end'] = self._io.pos()

                self._debug['guids']['end'] = self._io.pos()
                self._debug['frame_status']['start'] = self._io.pos()
                self.frame_status = self._io.read_bytes(4)
                self._debug['frame_status']['end'] = self._io.pos()
                self._debug['name_size']['start'] = self._io.pos()
                self.name_size = self._io.read_u4le()
                self._debug['name_size']['end'] = self._io.pos()
                self._debug['comm_size']['start'] = self._io.pos()
                self.comm_size = self._io.read_u4le()
                self._debug['comm_size']['end'] = self._io.pos()
                self._debug['view_info_size']['start'] = self._io.pos()
                self.view_info_size = self._io.read_u4le()
                self._debug['view_info_size']['end'] = self._io.pos()
                self._debug['spec_size']['start'] = self._io.pos()
                self.spec_size = self._io.read_u4le()
                self._debug['spec_size']['end'] = self._io.pos()
                self._debug['source_info_size']['start'] = self._io.pos()
                self.source_info_size = self._io.read_u4le()
                self._debug['source_info_size']['end'] = self._io.pos()
                self._debug['var_size']['start'] = self._io.pos()
                self.var_size = self._io.read_u4le()
                self._debug['var_size']['end'] = self._io.pos()
                self._debug['data_offset']['start'] = self._io.pos()
                self.data_offset = self._io.read_u4le()
                self._debug['data_offset']['end'] = self._io.pos()
                self._debug['data_size']['start'] = self._io.pos()
                self.data_size = self._io.read_u4le()
                self._debug['data_size']['end'] = self._io.pos()
                self._debug['title']['start'] = self._io.pos()
                self.title = (self._io.read_bytes(self.name_size)).decode(u"UTF-8")
                self._debug['title']['end'] = self._io.pos()
                self._debug['xml']['start'] = self._io.pos()
                self.xml = (self._io.read_bytes(self.comm_size)).decode(u"UTF-8")
                self._debug['xml']['end'] = self._io.pos()
                self._debug['struct_len']['start'] = self._io.pos()
                self.struct_len = self._io.read_u4le()
                self._debug['struct_len']['end'] = self._io.pos()
                self._debug['array_size']['start'] = self._io.pos()
                self.array_size = self._io.read_u8le()
                self._debug['array_size']['end'] = self._io.pos()
                self._debug['cell_size']['start'] = self._io.pos()
                self.cell_size = self._io.read_u4le()
                self._debug['cell_size']['end'] = self._io.pos()
                self._debug['n_dimensions']['start'] = self._io.pos()
                self.n_dimensions = self._io.read_u4le()
                self._debug['n_dimensions']['end'] = self._io.pos()
                self._debug['n_mesurands']['start'] = self._io.pos()
                self.n_mesurands = self._io.read_u4le()
                self._debug['n_mesurands']['end'] = self._io.pos()
                self._debug['dimensions']['start'] = self._io.pos()
                self.dimensions = [None] * (self.n_dimensions)
                for i in range(self.n_dimensions):
                    if not 'arr' in self._debug['dimensions']:
                        self._debug['dimensions']['arr'] = []
                    self._debug['dimensions']['arr'].append({'start': self._io.pos()})
                    _t_dimensions = NtMdt.Frame.FdMetaData.Calibration(self._io, self, self._root)
                    _t_dimensions._read()
                    self.dimensions[i] = _t_dimensions
                    self._debug['dimensions']['arr'][i]['end'] = self._io.pos()

                self._debug['dimensions']['end'] = self._io.pos()
                self._debug['mesurands']['start'] = self._io.pos()
                self.mesurands = [None] * (self.n_mesurands)
                for i in range(self.n_mesurands):
                    if not 'arr' in self._debug['mesurands']:
                        self._debug['mesurands']['arr'] = []
                    self._debug['mesurands']['arr'].append({'start': self._io.pos()})
                    _t_mesurands = NtMdt.Frame.FdMetaData.Calibration(self._io, self, self._root)
                    _t_mesurands._read()
                    self.mesurands[i] = _t_mesurands
                    self._debug['mesurands']['arr'][i]['end'] = self._io.pos()

                self._debug['mesurands']['end'] = self._io.pos()

            class Image(KaitaiStruct):
                SEQ_FIELDS = ["image"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['image']['start'] = self._io.pos()
                    self.image = []
                    i = 0
                    while not self._io.is_eof():
                        if not 'arr' in self._debug['image']:
                            self._debug['image']['arr'] = []
                        self._debug['image']['arr'].append({'start': self._io.pos()})
                        _t_image = NtMdt.Frame.FdMetaData.Image.Vec(self._io, self, self._root)
                        _t_image._read()
                        self.image.append(_t_image)
                        self._debug['image']['arr'][len(self.image) - 1]['end'] = self._io.pos()
                        i += 1

                    self._debug['image']['end'] = self._io.pos()

                class Vec(KaitaiStruct):
                    SEQ_FIELDS = ["items"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['items']['start'] = self._io.pos()
                        self.items = [None] * (self._parent._parent.n_mesurands)
                        for i in range(self._parent._parent.n_mesurands):
                            if not 'arr' in self._debug['items']:
                                self._debug['items']['arr'] = []
                            self._debug['items']['arr'].append({'start': self._io.pos()})
                            _on = self._parent._parent.mesurands[i].data_type
                            if _on == NtMdt.DataType.uint64:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_u8le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.uint8:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_u1()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.float32:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_f4le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.int8:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_s1()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.uint16:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_u2le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.int64:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_s8le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.uint32:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_u4le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.float64:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_f8le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.int16:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_s2le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            elif _on == NtMdt.DataType.int32:
                                if not 'arr' in self._debug['items']:
                                    self._debug['items']['arr'] = []
                                self._debug['items']['arr'].append({'start': self._io.pos()})
                                self.items[i] = self._io.read_s4le()
                                self._debug['items']['arr'][i]['end'] = self._io.pos()
                            self._debug['items']['arr'][i]['end'] = self._io.pos()

                        self._debug['items']['end'] = self._io.pos()



            class Calibration(KaitaiStruct):
                SEQ_FIELDS = ["len_tot", "len_struct", "len_name", "len_comment", "len_unit", "si_unit", "accuracy", "function_id_and_dimensions", "bias", "scale", "min_index", "max_index", "data_type", "len_author", "name", "comment", "unit", "author"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['len_tot']['start'] = self._io.pos()
                    self.len_tot = self._io.read_u4le()
                    self._debug['len_tot']['end'] = self._io.pos()
                    self._debug['len_struct']['start'] = self._io.pos()
                    self.len_struct = self._io.read_u4le()
                    self._debug['len_struct']['end'] = self._io.pos()
                    self._debug['len_name']['start'] = self._io.pos()
                    self.len_name = self._io.read_u4le()
                    self._debug['len_name']['end'] = self._io.pos()
                    self._debug['len_comment']['start'] = self._io.pos()
                    self.len_comment = self._io.read_u4le()
                    self._debug['len_comment']['end'] = self._io.pos()
                    self._debug['len_unit']['start'] = self._io.pos()
                    self.len_unit = self._io.read_u4le()
                    self._debug['len_unit']['end'] = self._io.pos()
                    self._debug['si_unit']['start'] = self._io.pos()
                    self.si_unit = self._io.read_u8le()
                    self._debug['si_unit']['end'] = self._io.pos()
                    self._debug['accuracy']['start'] = self._io.pos()
                    self.accuracy = self._io.read_f8le()
                    self._debug['accuracy']['end'] = self._io.pos()
                    self._debug['function_id_and_dimensions']['start'] = self._io.pos()
                    self.function_id_and_dimensions = self._io.read_u8le()
                    self._debug['function_id_and_dimensions']['end'] = self._io.pos()
                    self._debug['bias']['start'] = self._io.pos()
                    self.bias = self._io.read_f8le()
                    self._debug['bias']['end'] = self._io.pos()
                    self._debug['scale']['start'] = self._io.pos()
                    self.scale = self._io.read_f8le()
                    self._debug['scale']['end'] = self._io.pos()
                    self._debug['min_index']['start'] = self._io.pos()
                    self.min_index = self._io.read_u8le()
                    self._debug['min_index']['end'] = self._io.pos()
                    self._debug['max_index']['start'] = self._io.pos()
                    self.max_index = self._io.read_u8le()
                    self._debug['max_index']['end'] = self._io.pos()
                    self._debug['data_type']['start'] = self._io.pos()
                    self.data_type = KaitaiStream.resolve_enum(NtMdt.DataType, self._io.read_s4le())
                    self._debug['data_type']['end'] = self._io.pos()
                    self._debug['len_author']['start'] = self._io.pos()
                    self.len_author = self._io.read_u4le()
                    self._debug['len_author']['end'] = self._io.pos()
                    self._debug['name']['start'] = self._io.pos()
                    self.name = (self._io.read_bytes(self.len_name)).decode(u"utf-8")
                    self._debug['name']['end'] = self._io.pos()
                    self._debug['comment']['start'] = self._io.pos()
                    self.comment = (self._io.read_bytes(self.len_comment)).decode(u"utf-8")
                    self._debug['comment']['end'] = self._io.pos()
                    self._debug['unit']['start'] = self._io.pos()
                    self.unit = (self._io.read_bytes(self.len_unit)).decode(u"utf-8")
                    self._debug['unit']['end'] = self._io.pos()
                    self._debug['author']['start'] = self._io.pos()
                    self.author = (self._io.read_bytes(self.len_author)).decode(u"utf-8")
                    self._debug['author']['end'] = self._io.pos()

                @property
                def count(self):
                    if hasattr(self, '_m_count'):
                        return self._m_count if hasattr(self, '_m_count') else None

                    self._m_count = ((self.max_index - self.min_index) + 1)
                    return self._m_count if hasattr(self, '_m_count') else None


            @property
            def image(self):
                if hasattr(self, '_m_image'):
                    return self._m_image if hasattr(self, '_m_image') else None

                _pos = self._io.pos()
                self._io.seek(self.data_offset)
                self._debug['_m_image']['start'] = self._io.pos()
                self._raw__m_image = self._io.read_bytes(self.data_size)
                _io__raw__m_image = KaitaiStream(BytesIO(self._raw__m_image))
                self._m_image = NtMdt.Frame.FdMetaData.Image(_io__raw__m_image, self, self._root)
                self._m_image._read()
                self._debug['_m_image']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_image if hasattr(self, '_m_image') else None


        class FdSpectroscopy(KaitaiStruct):
            SEQ_FIELDS = ["vars", "fm_mode", "fm_xres", "fm_yres", "dots", "data", "title", "xml"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['vars']['start'] = self._io.pos()
                self._raw_vars = self._io.read_bytes(self._parent.var_size)
                _io__raw_vars = KaitaiStream(BytesIO(self._raw_vars))
                self.vars = NtMdt.Frame.FdSpectroscopy.Vars(_io__raw_vars, self, self._root)
                self.vars._read()
                self._debug['vars']['end'] = self._io.pos()
                self._debug['fm_mode']['start'] = self._io.pos()
                self.fm_mode = self._io.read_u2le()
                self._debug['fm_mode']['end'] = self._io.pos()
                self._debug['fm_xres']['start'] = self._io.pos()
                self.fm_xres = self._io.read_u2le()
                self._debug['fm_xres']['end'] = self._io.pos()
                self._debug['fm_yres']['start'] = self._io.pos()
                self.fm_yres = self._io.read_u2le()
                self._debug['fm_yres']['end'] = self._io.pos()
                self._debug['dots']['start'] = self._io.pos()
                self.dots = NtMdt.Frame.Dots(self._io, self, self._root)
                self.dots._read()
                self._debug['dots']['end'] = self._io.pos()
                self._debug['data']['start'] = self._io.pos()
                self.data = [None] * ((self.fm_xres * self.fm_yres))
                for i in range((self.fm_xres * self.fm_yres)):
                    if not 'arr' in self._debug['data']:
                        self._debug['data']['arr'] = []
                    self._debug['data']['arr'].append({'start': self._io.pos()})
                    self.data[i] = self._io.read_s2le()
                    self._debug['data']['arr'][i]['end'] = self._io.pos()

                self._debug['data']['end'] = self._io.pos()
                self._debug['title']['start'] = self._io.pos()
                self.title = NtMdt.Title(self._io, self, self._root)
                self.title._read()
                self._debug['title']['end'] = self._io.pos()
                self._debug['xml']['start'] = self._io.pos()
                self.xml = NtMdt.Xml(self._io, self, self._root)
                self.xml._read()
                self._debug['xml']['end'] = self._io.pos()

            class Vars(KaitaiStruct):
                SEQ_FIELDS = ["x_scale", "y_scale", "z_scale", "sp_mode", "sp_filter", "u_begin", "u_end", "z_up", "z_down", "sp_averaging", "sp_repeat", "sp_back", "sp_4nx", "sp_osc", "sp_n4", "sp_4x0", "sp_4xr", "sp_4u", "sp_4i", "sp_nx"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['x_scale']['start'] = self._io.pos()
                    self.x_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.x_scale._read()
                    self._debug['x_scale']['end'] = self._io.pos()
                    self._debug['y_scale']['start'] = self._io.pos()
                    self.y_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.y_scale._read()
                    self._debug['y_scale']['end'] = self._io.pos()
                    self._debug['z_scale']['start'] = self._io.pos()
                    self.z_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.z_scale._read()
                    self._debug['z_scale']['end'] = self._io.pos()
                    self._debug['sp_mode']['start'] = self._io.pos()
                    self.sp_mode = self._io.read_u2le()
                    self._debug['sp_mode']['end'] = self._io.pos()
                    self._debug['sp_filter']['start'] = self._io.pos()
                    self.sp_filter = self._io.read_u2le()
                    self._debug['sp_filter']['end'] = self._io.pos()
                    self._debug['u_begin']['start'] = self._io.pos()
                    self.u_begin = self._io.read_f4le()
                    self._debug['u_begin']['end'] = self._io.pos()
                    self._debug['u_end']['start'] = self._io.pos()
                    self.u_end = self._io.read_f4le()
                    self._debug['u_end']['end'] = self._io.pos()
                    self._debug['z_up']['start'] = self._io.pos()
                    self.z_up = self._io.read_s2le()
                    self._debug['z_up']['end'] = self._io.pos()
                    self._debug['z_down']['start'] = self._io.pos()
                    self.z_down = self._io.read_s2le()
                    self._debug['z_down']['end'] = self._io.pos()
                    self._debug['sp_averaging']['start'] = self._io.pos()
                    self.sp_averaging = self._io.read_u2le()
                    self._debug['sp_averaging']['end'] = self._io.pos()
                    self._debug['sp_repeat']['start'] = self._io.pos()
                    self.sp_repeat = self._io.read_u1()
                    self._debug['sp_repeat']['end'] = self._io.pos()
                    self._debug['sp_back']['start'] = self._io.pos()
                    self.sp_back = self._io.read_u1()
                    self._debug['sp_back']['end'] = self._io.pos()
                    self._debug['sp_4nx']['start'] = self._io.pos()
                    self.sp_4nx = self._io.read_s2le()
                    self._debug['sp_4nx']['end'] = self._io.pos()
                    self._debug['sp_osc']['start'] = self._io.pos()
                    self.sp_osc = self._io.read_u1()
                    self._debug['sp_osc']['end'] = self._io.pos()
                    self._debug['sp_n4']['start'] = self._io.pos()
                    self.sp_n4 = self._io.read_u1()
                    self._debug['sp_n4']['end'] = self._io.pos()
                    self._debug['sp_4x0']['start'] = self._io.pos()
                    self.sp_4x0 = self._io.read_f4le()
                    self._debug['sp_4x0']['end'] = self._io.pos()
                    self._debug['sp_4xr']['start'] = self._io.pos()
                    self.sp_4xr = self._io.read_f4le()
                    self._debug['sp_4xr']['end'] = self._io.pos()
                    self._debug['sp_4u']['start'] = self._io.pos()
                    self.sp_4u = self._io.read_s2le()
                    self._debug['sp_4u']['end'] = self._io.pos()
                    self._debug['sp_4i']['start'] = self._io.pos()
                    self.sp_4i = self._io.read_s2le()
                    self._debug['sp_4i']['end'] = self._io.pos()
                    self._debug['sp_nx']['start'] = self._io.pos()
                    self.sp_nx = self._io.read_s2le()
                    self._debug['sp_nx']['end'] = self._io.pos()



        class DateTime(KaitaiStruct):
            SEQ_FIELDS = ["date", "time"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['date']['start'] = self._io.pos()
                self.date = NtMdt.Frame.DateTime.Date(self._io, self, self._root)
                self.date._read()
                self._debug['date']['end'] = self._io.pos()
                self._debug['time']['start'] = self._io.pos()
                self.time = NtMdt.Frame.DateTime.Time(self._io, self, self._root)
                self.time._read()
                self._debug['time']['end'] = self._io.pos()

            class Date(KaitaiStruct):
                SEQ_FIELDS = ["year", "month", "day"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['year']['start'] = self._io.pos()
                    self.year = self._io.read_u2le()
                    self._debug['year']['end'] = self._io.pos()
                    self._debug['month']['start'] = self._io.pos()
                    self.month = self._io.read_u2le()
                    self._debug['month']['end'] = self._io.pos()
                    self._debug['day']['start'] = self._io.pos()
                    self.day = self._io.read_u2le()
                    self._debug['day']['end'] = self._io.pos()


            class Time(KaitaiStruct):
                SEQ_FIELDS = ["hour", "min", "sec"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['hour']['start'] = self._io.pos()
                    self.hour = self._io.read_u2le()
                    self._debug['hour']['end'] = self._io.pos()
                    self._debug['min']['start'] = self._io.pos()
                    self.min = self._io.read_u2le()
                    self._debug['min']['end'] = self._io.pos()
                    self._debug['sec']['start'] = self._io.pos()
                    self.sec = self._io.read_u2le()
                    self._debug['sec']['end'] = self._io.pos()



        class AxisScale(KaitaiStruct):
            SEQ_FIELDS = ["offset", "step", "unit"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['offset']['start'] = self._io.pos()
                self.offset = self._io.read_f4le()
                self._debug['offset']['end'] = self._io.pos()
                self._debug['step']['start'] = self._io.pos()
                self.step = self._io.read_f4le()
                self._debug['step']['end'] = self._io.pos()
                self._debug['unit']['start'] = self._io.pos()
                self.unit = KaitaiStream.resolve_enum(NtMdt.Unit, self._io.read_s2le())
                self._debug['unit']['end'] = self._io.pos()


        class FdScanned(KaitaiStruct):

            class Mode(Enum):
                stm = 0
                afm = 1
                unknown2 = 2
                unknown3 = 3
                unknown4 = 4

            class InputSignal(Enum):
                extension_slot = 0
                bias_v = 1
                ground = 2

            class LiftMode(Enum):
                step = 0
                fine = 1
                slope = 2
            SEQ_FIELDS = ["vars", "orig_format", "tune", "feedback_gain", "dac_scale", "overscan", "fm_mode", "fm_xres", "fm_yres", "dots", "image", "title", "xml"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['vars']['start'] = self._io.pos()
                self._raw_vars = self._io.read_bytes(self._parent.var_size)
                _io__raw_vars = KaitaiStream(BytesIO(self._raw_vars))
                self.vars = NtMdt.Frame.FdScanned.Vars(_io__raw_vars, self, self._root)
                self.vars._read()
                self._debug['vars']['end'] = self._io.pos()
                if False:
                    self._debug['orig_format']['start'] = self._io.pos()
                    self.orig_format = self._io.read_u4le()
                    self._debug['orig_format']['end'] = self._io.pos()

                if False:
                    self._debug['tune']['start'] = self._io.pos()
                    self.tune = KaitaiStream.resolve_enum(NtMdt.Frame.FdScanned.LiftMode, self._io.read_u4le())
                    self._debug['tune']['end'] = self._io.pos()

                if False:
                    self._debug['feedback_gain']['start'] = self._io.pos()
                    self.feedback_gain = self._io.read_f8le()
                    self._debug['feedback_gain']['end'] = self._io.pos()

                if False:
                    self._debug['dac_scale']['start'] = self._io.pos()
                    self.dac_scale = self._io.read_s4le()
                    self._debug['dac_scale']['end'] = self._io.pos()

                if False:
                    self._debug['overscan']['start'] = self._io.pos()
                    self.overscan = self._io.read_s4le()
                    self._debug['overscan']['end'] = self._io.pos()

                self._debug['fm_mode']['start'] = self._io.pos()
                self.fm_mode = self._io.read_u2le()
                self._debug['fm_mode']['end'] = self._io.pos()
                self._debug['fm_xres']['start'] = self._io.pos()
                self.fm_xres = self._io.read_u2le()
                self._debug['fm_xres']['end'] = self._io.pos()
                self._debug['fm_yres']['start'] = self._io.pos()
                self.fm_yres = self._io.read_u2le()
                self._debug['fm_yres']['end'] = self._io.pos()
                self._debug['dots']['start'] = self._io.pos()
                self.dots = NtMdt.Frame.Dots(self._io, self, self._root)
                self.dots._read()
                self._debug['dots']['end'] = self._io.pos()
                self._debug['image']['start'] = self._io.pos()
                self.image = [None] * ((self.fm_xres * self.fm_yres))
                for i in range((self.fm_xres * self.fm_yres)):
                    if not 'arr' in self._debug['image']:
                        self._debug['image']['arr'] = []
                    self._debug['image']['arr'].append({'start': self._io.pos()})
                    self.image[i] = self._io.read_s2le()
                    self._debug['image']['arr'][i]['end'] = self._io.pos()

                self._debug['image']['end'] = self._io.pos()
                self._debug['title']['start'] = self._io.pos()
                self.title = NtMdt.Title(self._io, self, self._root)
                self.title._read()
                self._debug['title']['end'] = self._io.pos()
                self._debug['xml']['start'] = self._io.pos()
                self.xml = NtMdt.Xml(self._io, self, self._root)
                self.xml._read()
                self._debug['xml']['end'] = self._io.pos()

            class Vars(KaitaiStruct):
                SEQ_FIELDS = ["x_scale", "y_scale", "z_scale", "channel_index", "mode", "xres", "yres", "ndacq", "step_length", "adt", "adc_gain_amp_log10", "adc_index", "input_signal_or_version", "substr_plane_order_or_pass_num", "scan_dir", "power_of_2", "velocity", "setpoint", "bias_voltage", "draw", "reserved", "xoff", "yoff", "nl_corr"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['x_scale']['start'] = self._io.pos()
                    self.x_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.x_scale._read()
                    self._debug['x_scale']['end'] = self._io.pos()
                    self._debug['y_scale']['start'] = self._io.pos()
                    self.y_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.y_scale._read()
                    self._debug['y_scale']['end'] = self._io.pos()
                    self._debug['z_scale']['start'] = self._io.pos()
                    self.z_scale = NtMdt.Frame.AxisScale(self._io, self, self._root)
                    self.z_scale._read()
                    self._debug['z_scale']['end'] = self._io.pos()
                    self._debug['channel_index']['start'] = self._io.pos()
                    self.channel_index = KaitaiStream.resolve_enum(NtMdt.AdcMode, self._io.read_u1())
                    self._debug['channel_index']['end'] = self._io.pos()
                    self._debug['mode']['start'] = self._io.pos()
                    self.mode = KaitaiStream.resolve_enum(NtMdt.Frame.FdScanned.Mode, self._io.read_u1())
                    self._debug['mode']['end'] = self._io.pos()
                    self._debug['xres']['start'] = self._io.pos()
                    self.xres = self._io.read_u2le()
                    self._debug['xres']['end'] = self._io.pos()
                    self._debug['yres']['start'] = self._io.pos()
                    self.yres = self._io.read_u2le()
                    self._debug['yres']['end'] = self._io.pos()
                    self._debug['ndacq']['start'] = self._io.pos()
                    self.ndacq = self._io.read_u2le()
                    self._debug['ndacq']['end'] = self._io.pos()
                    self._debug['step_length']['start'] = self._io.pos()
                    self.step_length = self._io.read_f4le()
                    self._debug['step_length']['end'] = self._io.pos()
                    self._debug['adt']['start'] = self._io.pos()
                    self.adt = self._io.read_u2le()
                    self._debug['adt']['end'] = self._io.pos()
                    self._debug['adc_gain_amp_log10']['start'] = self._io.pos()
                    self.adc_gain_amp_log10 = self._io.read_u1()
                    self._debug['adc_gain_amp_log10']['end'] = self._io.pos()
                    self._debug['adc_index']['start'] = self._io.pos()
                    self.adc_index = self._io.read_u1()
                    self._debug['adc_index']['end'] = self._io.pos()
                    self._debug['input_signal_or_version']['start'] = self._io.pos()
                    self.input_signal_or_version = self._io.read_u1()
                    self._debug['input_signal_or_version']['end'] = self._io.pos()
                    self._debug['substr_plane_order_or_pass_num']['start'] = self._io.pos()
                    self.substr_plane_order_or_pass_num = self._io.read_u1()
                    self._debug['substr_plane_order_or_pass_num']['end'] = self._io.pos()
                    self._debug['scan_dir']['start'] = self._io.pos()
                    self.scan_dir = NtMdt.Frame.FdScanned.ScanDir(self._io, self, self._root)
                    self.scan_dir._read()
                    self._debug['scan_dir']['end'] = self._io.pos()
                    self._debug['power_of_2']['start'] = self._io.pos()
                    self.power_of_2 = self._io.read_u1()
                    self._debug['power_of_2']['end'] = self._io.pos()
                    self._debug['velocity']['start'] = self._io.pos()
                    self.velocity = self._io.read_f4le()
                    self._debug['velocity']['end'] = self._io.pos()
                    self._debug['setpoint']['start'] = self._io.pos()
                    self.setpoint = self._io.read_f4le()
                    self._debug['setpoint']['end'] = self._io.pos()
                    self._debug['bias_voltage']['start'] = self._io.pos()
                    self.bias_voltage = self._io.read_f4le()
                    self._debug['bias_voltage']['end'] = self._io.pos()
                    self._debug['draw']['start'] = self._io.pos()
                    self.draw = self._io.read_u1()
                    self._debug['draw']['end'] = self._io.pos()
                    self._debug['reserved']['start'] = self._io.pos()
                    self.reserved = self._io.read_u1()
                    self._debug['reserved']['end'] = self._io.pos()
                    self._debug['xoff']['start'] = self._io.pos()
                    self.xoff = self._io.read_s4le()
                    self._debug['xoff']['end'] = self._io.pos()
                    self._debug['yoff']['start'] = self._io.pos()
                    self.yoff = self._io.read_s4le()
                    self._debug['yoff']['end'] = self._io.pos()
                    self._debug['nl_corr']['start'] = self._io.pos()
                    self.nl_corr = self._io.read_u1()
                    self._debug['nl_corr']['end'] = self._io.pos()


            class Dot(KaitaiStruct):
                SEQ_FIELDS = ["x", "y"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['x']['start'] = self._io.pos()
                    self.x = self._io.read_s2le()
                    self._debug['x']['end'] = self._io.pos()
                    self._debug['y']['start'] = self._io.pos()
                    self.y = self._io.read_s2le()
                    self._debug['y']['end'] = self._io.pos()


            class ScanDir(KaitaiStruct):
                SEQ_FIELDS = ["unkn", "double_pass", "bottom", "left", "horizontal"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['unkn']['start'] = self._io.pos()
                    self.unkn = self._io.read_bits_int_be(4)
                    self._debug['unkn']['end'] = self._io.pos()
                    self._debug['double_pass']['start'] = self._io.pos()
                    self.double_pass = self._io.read_bits_int_be(1) != 0
                    self._debug['double_pass']['end'] = self._io.pos()
                    self._debug['bottom']['start'] = self._io.pos()
                    self.bottom = self._io.read_bits_int_be(1) != 0
                    self._debug['bottom']['end'] = self._io.pos()
                    self._debug['left']['start'] = self._io.pos()
                    self.left = self._io.read_bits_int_be(1) != 0
                    self._debug['left']['end'] = self._io.pos()
                    self._debug['horizontal']['start'] = self._io.pos()
                    self.horizontal = self._io.read_bits_int_be(1) != 0
                    self._debug['horizontal']['end'] = self._io.pos()




    class Version(KaitaiStruct):
        SEQ_FIELDS = ["minor", "major"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['minor']['start'] = self._io.pos()
            self.minor = self._io.read_u1()
            self._debug['minor']['end'] = self._io.pos()
            self._debug['major']['start'] = self._io.pos()
            self.major = self._io.read_u1()
            self._debug['major']['end'] = self._io.pos()


    class Xml(KaitaiStruct):
        SEQ_FIELDS = ["xml_len", "xml"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['xml_len']['start'] = self._io.pos()
            self.xml_len = self._io.read_u4le()
            self._debug['xml_len']['end'] = self._io.pos()
            self._debug['xml']['start'] = self._io.pos()
            self.xml = (self._io.read_bytes(self.xml_len)).decode(u"UTF-16LE")
            self._debug['xml']['end'] = self._io.pos()


    class Title(KaitaiStruct):
        SEQ_FIELDS = ["title_len", "title"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['title_len']['start'] = self._io.pos()
            self.title_len = self._io.read_u4le()
            self._debug['title_len']['end'] = self._io.pos()
            self._debug['title']['start'] = self._io.pos()
            self.title = (self._io.read_bytes(self.title_len)).decode(u"cp1251")
            self._debug['title']['end'] = self._io.pos()



